// GoBoard.cc
// GoBoard for use with goVRec
// Andreas Balser

#define MAXSIZE 19

//for counting
#define NOBODY       0
#define WH           1
#define BL           2
#define NO_POINT     3
#define SET_LATER    4

class GoBoard {
 private:
  int size, width,stonerad,xstart,ystart;
  int ko_x, ko_y;
  int s_width, s_height;
  bool nowQuiet;
  bool coordinates;
  bool isStar[MAXSIZE][MAXSIZE];
  bool isOccupied[MAXSIZE][MAXSIZE];
  bool isWhite[MAXSIZE][MAXSIZE];
  bool ch1[MAXSIZE][MAXSIZE];
  bool ch2[MAXSIZE][MAXSIZE];
  bool ch3[MAXSIZE][MAXSIZE];
  bool ch4[MAXSIZE][MAXSIZE];
  int terr[MAXSIZE][MAXSIZE];

  GoGraphic* graph;

  void drawStone(int, int, bool);
  void markStar(int,int);
  void intRemStone(int,int);
  bool dead(int,int);
  int remove(int,int);
  void rem(int, int);
  void clear(int);
  void markO(int, int);
  bool koAlone(int, int);

  //counting routines
  int markBound(int,int);
  void markTerr(int,int,int);
  int belongsWho();
  void drawHalfStone(int,int,int);
  int countInsideLibs(int,int, bool);
  void makeNeutral(int,int);
  void computeNeutrals();
  void countremove(int,int,int&,int&);
  int changecount(int,int, int whose = NOBODY);
  bool isEye(int,int);
  void doChanges();
  bool tryChange(int&,int&);
  void drawTerritory();

public:
  GoBoard (GoGraphic*, int, int);
  void GoBoard::setProperties(int s, int xs, int ys, int w, bool);
  void GoBoard::resize(int xs, int ys, int w, bool coord);
  void redraw();
  void draw ();
  int setStone(int,int,bool, int&, bool quiet = false);
  bool removeStone(int,int, int, int&, bool quiet = false);
  int fill(int x, int y, bool White);
  void markStone(int,int,int, bool quiet = false);
  void initHandi(int[], int[], int, bool, bool quiet = false);
  bool performCountAction(int,int,int&,int&);
  void initialCount(int&, int&);
  void resetTerr();
  bool isOcc(int,int);
};

GoBoard::GoBoard(GoGraphic* gr, int w, int h){
  graph = gr;
  s_width = w;
  s_height= h;
}

void GoBoard::setProperties(int s, int xs, int ys, int w, bool coord){
  
  if (s > MAXSIZE) {
    s = MAXSIZE;
  }
  size = s;
  for(int i=0; i<size; i++)
    for(int j=0; j<size; j++){
      isStar[i][j] = false;
      isOccupied[i][j] = false;
      ch1[i][j] = false;
      ch2[i][j] = false;
      terr[i][j] = NOBODY;

  }
  if (size == 19||size==13)
    isStar[3][3]=isStar[9][3]=
      isStar[3][9]=isStar[9][9]= true;
  
  if(size==19)
    isStar[15][3]=isStar[15][9]=isStar[3][15]
      =isStar[9][15]=isStar[15][15]=true;
  if(size==9)
    isStar[2][2]=isStar[6][6]=isStar[2][6]=isStar[6][2]=true;

  ko_x=NORMALMARK;
  ko_y=-1;
  width=w;
  xstart=xs;
  ystart=ys;
  stonerad = width / 2;
  //  printf("%d\n",width);
  nowQuiet = false;
  coordinates = coord;
  //  ko_x=-1;
  //  last_x=-1;
  
  //  draw();
}


void GoBoard::resize(int xs, int ys, int w, bool coord){
  width=w;
  xstart=xs;
  ystart=ys;
  stonerad = width / 2;
  coordinates=coord;
}

void GoBoard::redraw(){
  graph->backRectangle(0,windowheight-windowwidth+1,
		       windowwidth, windowwidth);
  draw();
}

void GoBoard::draw(){
  int i,j;
  if(coordinates){
    char name[3];
    int xcoord;
    name[1]=name[2]=0;
    for(i=0; i<size; i++){
      name[0]='A'+i;
      if(i>7) name[0]++;
      name[1]=0;
      graph->dSString(xstart+i*width-s_width/2, windowheight, name);
#if defined SMALLFONT	
      xcoord=2;
#else
      xcoord=3+s_width/2;
#endif
      if(i<9) {
	name[0]=i+'1';
      }
      else {
	xcoord=2;
	name[0]='1';
#if defined SMALLFONT	
	name[0]=(i-9)+'0';
	graph->dOneLine(2, ystart+(size-1-i)*width+s_height/2);
#else
	name[1]=(i-9)+'0';
#endif
      }
      graph->dSString(xcoord, ystart+(size-1-i)*width+s_height/2, name);
    }
  }
  for (i = 0; i < size; i++){
    graph->line( xstart+ i*width, ystart, 
		 xstart + i*width, ystart + (size-1)*width);
    graph->line( xstart, ystart+ i*width, 
		 xstart + (size-1)*width, ystart + i*width);
  }
  graph->line( xstart-1, ystart-1, 
	       xstart+1+(size-1)*width, ystart-1);
  graph->line( xstart-1, ystart-1, 
	       xstart-1, ystart+1+(size-1)*width);
  graph->line( xstart-1, ystart+1+(size-1)*width, 
	       xstart+1+(size-1)*width, ystart+1+(size-1)*width);
  graph->line( xstart+1+(size-1)*width, ystart-1, 
	       xstart+1+(size-1)*width, ystart+1+(size-1)*width);
  
  for(i=0; i<size; i++)
    for(j=0; j<size; j++){
      if(isOccupied[i][j]) drawStone(i,j,isWhite[i][j]);
      else if(isStar[i][j]) markStar(i,j);
      if(terr[i][j] != NOBODY) drawHalfStone(i,j,terr[i][j]);
    }
  nowQuiet = false;
  markStone(-2,-1,REDRAWMARK);

}

void GoBoard::drawTerritory(){
  int i,j;
  for(i=0; i<size; i++)
    for(j=0; j<size; j++){
      if(terr[i][j] != NOBODY) drawHalfStone(i,j,terr[i][j]);
    }
  //  markStone(-2,-1,REDRAW);
}

int GoBoard::setStone(int x, int y, bool White, int& addKilled, 
		      bool quiet){
  nowQuiet = quiet;
  int ret = OK;
  int killed=0;
  if(isOccupied[x][y]) return false;
  if(x == ko_x && y == ko_y && x != -1) return false;
  ko_x=NORMALMARK;
  
  isOccupied[x][y]=true;
  isWhite[x][y]=White;

  for(int i=0; i<size; i++){
    for(int j=0; j<size; j++)
      ch1[i][j]=false;
  }

  if(x!=0 && isOccupied[x-1][y] && isWhite[x-1][y]!=White){
    if(dead(x-1, y)) {killed+=remove(x-1,y);ret|=L;}
  }
  if(x!=size-1 && !ch1[x+1][y] && isOccupied[x+1][y] && isWhite[x+1][y]!=White)
    if(dead(x+1,y)) {killed+=remove(x+1,y); ret|=R;}
  if(y!=0 && !ch1[x][y-1] && isOccupied[x][y-1] && isWhite[x][y-1]!=White){
    if(dead(x,y-1)) {killed+=remove(x,y-1); ret|=U;}
  }
  if(y!=size-1 && !ch1[x][y+1] && isOccupied[x][y+1] && isWhite[x][y+1]!=White)
    if(dead(x,y+1)) {killed+=remove(x,y+1); ret|=D;}

  if(dead(x,y)){
    isOccupied[x][y]=false;
    return NO;
  }

  if(!quiet) drawStone(x,y,White);

  if(killed == 1 && koAlone(x,y)){
    if((ret & L) != 0) {ko_x=x-1; ko_y=y;}
    else if((ret & R) != 0) {ko_x=x+1; ko_y=y;}
    else if((ret & U) != 0) {ko_x=x; ko_y=y-1;}
    else if((ret & D) != 0) {ko_x=x; ko_y=y+1;}
    ret|=KO;
  }

  markStone (x,y,ko_x, quiet);

  addKilled+= killed;
  return (ret+killed*MULT);
}

void GoBoard::drawStone(int x, int y, bool White){
  graph->fCircle( xstart+x*width - stonerad, ystart+width*y-stonerad, 
		    2*stonerad, White);
  graph->circle(xstart+x*width - stonerad, ystart+width*y-stonerad, 
		2*stonerad, BLACK);
}

void GoBoard::markStar(int x, int y){
  graph->point( xstart+ x*width -1, ystart+y*width -1);
  graph->point( xstart+ x*width +1, ystart+y*width -1);
  graph->point( xstart+ x*width -1, ystart+y*width +1);
  graph->point( xstart+ x*width +1, ystart+y*width +1);

}

void GoBoard::intRemStone(int x, int y){
    isOccupied[x][y]=false;
}

bool GoBoard::removeStone(int x, int y, int what, int& subKilled, 
			  bool quiet){
  int sub=0;
  nowQuiet = quiet;
  ko_x = NORMALMARK;
  if ((what & KO)!= 0) 
    if((what & U) != 0) {ko_x = x; ko_y = y-1;}
    else if((what & D) != 0) {ko_x = x; ko_y = y+1;}
    else if((what & L) != 0) {ko_x = x-1; ko_y = y;}
    else if((what & R) != 0) {ko_x = x+1; ko_y = y;}

  if(x<0 || y<0) return false;
  if (!isOccupied[x][y]){
    fprintf(stderr,"not occupied!?\n");
    return false;
  }
  // this section is for restoring stones which were killed when the now 
  // removed stone was placed.

  if((what & U) != 0) sub+=fill(x, y-1, !isWhite[x][y]);
  if((what & D) != 0) sub+=fill(x, y+1, !isWhite[x][y]);
  if((what & L) != 0) sub+=fill(x-1, y, !isWhite[x][y]);
  if((what & R) != 0) sub+=fill(x+1, y, !isWhite[x][y]);

  subKilled-=sub;

  isOccupied[x][y]=false;

  if(!nowQuiet){
    graph->fCircle(xstart+x*width - stonerad, 
		   ystart+width*y-stonerad, 2*stonerad, true,true);

    graph->circle(xstart+x*width - stonerad, 
		  ystart+width*y-stonerad, 2*stonerad, true,true);

  if(x!=0 && x!=size-1)
    graph->line( xstart+x*width-stonerad, ystart+y*width,
	      xstart+x*width+stonerad, ystart+y*width);
  else if (x==0){
    graph->line( xstart+x*width, ystart+y*width,
	      xstart+x*width+stonerad, ystart+y*width);
    if(y!=0 && y!=size-1) graph->line( xstart -1, 
		       ystart+y*width-stonerad, xstart-1, ystart+y*width+stonerad);
    else if (y==0)
      graph->line( xstart -1, ystart-1, 
		xstart-1, ystart+y*width+stonerad);
    else 
      graph->line( xstart -1, ystart+y*width-stonerad, 
		xstart-1, ystart+y*width+1);

  }
  else{ // x=size-1
    graph->line( xstart+x*width-stonerad, ystart+y*width,
	      xstart+x*width, ystart+y*width);
    if(y!=0 && y!=size-1) 
      graph->line( xstart +x*width+1, 
		ystart+y*width-stonerad, xstart+x*width+1, 
		ystart+y*width+stonerad);
    else if (y==0)
      graph->line( xstart +x*width+1, ystart-1, 
		xstart+x*width+1, ystart+y*width+stonerad);
    else 
      graph->line( xstart +x*width+1, 
		ystart+y*width-stonerad, 
		xstart+x*width+1, ystart+y*width+1);
  
  }

  if(y!=0 && y!=size-1)
    graph->line( xstart+x*width, ystart+y*width-stonerad,
	      xstart+x*width, ystart+y*width+stonerad);
  else if (y==0){
    graph->line( xstart+x*width, ystart+y*width,
	      xstart+x*width, ystart+y*width+stonerad);
    if(x!=0 && x!=size-1)
      graph->line(xstart+x*width-stonerad,ystart-1,
		xstart+x*width+stonerad, ystart-1);
    else if (x==0)
      graph->line(xstart-1,ystart-1,
		xstart+x*width+stonerad, ystart-1);
    else
      graph->line(xstart+x*width-stonerad,ystart-1,
		xstart+x*width+1, ystart-1);


  }
  else { // y=size-1
    graph->line( xstart+x*width, ystart+y*width-stonerad,
	      xstart+x*width, ystart+y*width);

    if(x!=0 && x!=size-1)
      graph->line(xstart+x*width-stonerad,ystart+y*width+1,
		xstart+x*width+stonerad, ystart+y*width+1);
    else if (x==0)
      graph->line(xstart-1,ystart+y*width+1,
		xstart+x*width+stonerad, ystart+y*width+1);
    else
      graph->line(xstart+x*width-stonerad,ystart+y*width+1,
		xstart+x*width+1, ystart+y*width+1);

  }

  if(isStar[x][y]) markStar(x,y);
  if(x!=0 && isOccupied[x-1][y]) drawStone(x-1,y,isWhite[x-1][y]);
  if(y!=0 && isOccupied[x][y-1]) drawStone(x,y-1,isWhite[x][y-1]);
  if(x!=size-1 && isOccupied[x+1][y]) drawStone(x+1,y,isWhite[x+1][y]);
  if(y!=size-1 && isOccupied[x][y+1]) drawStone(x,y+1,isWhite[x][y+1]);
  }
  return true;
}

bool GoBoard::dead(int x, int y){
  //  fprintf(stderr,"dead for %d,%d\n",x,y);
  ch1[x][y] = true;
  bool interm=true;
  if(x!=0 && !ch1[x-1][y]&&isOccupied[x-1][y] && isWhite[x-1][y]==isWhite[x][y])
    interm &= dead(x-1,y);
  if(x!=size-1 && !ch1[x+1][y]&&isOccupied[x+1][y] && isWhite[x+1][y]==isWhite[x][y]) 
    interm &= dead(x+1,y);
  if(y!=0 && !ch1[x][y-1]&&isOccupied[x][y-1] && isWhite[x][y-1]==isWhite[x][y])
    interm &= dead(x,y-1);
  if(y!=size-1 && !ch1[x][y+1]&&isOccupied[x][y+1] && isWhite[x][y+1]==isWhite[x][y]) 
    interm &= dead(x,y+1);

  if(!interm) return false;
  //  fprintf(stderr,"still");
  if(x!=0 && !isOccupied[x-1][y]) {
    //  fprintf(stderr,"1\n");
    return false;
  }
  if(x!=size-1 && !isOccupied[x+1][y]) {
    //    fprintf(stderr,"2\n");
    return false;
  }
  if(y!=0 && !isOccupied[x][y-1]) {
    //    fprintf(stderr,"3\n");
    return false;
  }
  if(y!=size-1 && !isOccupied[x][y+1]) {
    //  fprintf(stderr,"4\n");
    return false;
  }

  return true;
}

int GoBoard::remove(int x, int y){
  int removed=0;
  int dummy=0;
  for(int i=0; i<size; i++)
    for(int j=0; j<size; j++)
      ch2[i][j]=false;
  rem(x,y);
  for(int i=0; i<size; i++)
    for(int j=0; j<size; j++)
      if(ch2[i][j]) {
	removed++;
	removeStone(i,j,OK,dummy, nowQuiet);
      }
  return removed;
}

void GoBoard::rem(int x, int y){
  ch2[x][y]=true;
  if(x!=0 && !ch2[x-1][y]&&isOccupied[x-1][y] && isWhite[x-1][y]==isWhite[x][y])
    rem(x-1,y);
  if(x!=size-1 && !ch2[x+1][y]&&isOccupied[x+1][y] && isWhite[x+1][y]==isWhite[x][y]) 
    rem(x+1,y);
  if(y!=0 && !ch2[x][y-1]&&isOccupied[x][y-1] && isWhite[x][y-1]==isWhite[x][y])
    rem(x,y-1);
  if(y!=size-1 && !ch2[x][y+1]&&isOccupied[x][y+1] && isWhite[x][y+1]==isWhite[x][y]) 
    rem(x,y+1);
  
}

//status >= 0 for KO, PLUS for normal, REPLACE for REPLACE
void GoBoard::markStone(int x, int y, int stat, bool quiet){
  static int last_x=-1, last_y=-1, last_stat=PLUS;
  static bool last_draw = false;
  if(stat==PASSMARKW || stat==PASSMARKB) ko_x=NORMALMARK;
  if(last_stat == PASS) fprintf(stderr, "PASS!\n");
  nowQuiet = quiet;
  if(stat!=REDRAWMARK) {
    if(last_stat == PASSMARKW || last_stat == PASSMARKB){
      graph->emptyRectangle(windowwidth-2*FONTWIDTH-1, 
			    4*BUTTONSIZE-3-FONTHEIGHT, 
			    2*FONTWIDTH,
			    FONTHEIGHT);
    }
    else
      if((isOccupied[last_x][last_y]) && last_draw) {
	drawStone(last_x,last_y, isWhite[last_x][last_y]);
      }
      else if(!isOccupied[last_x][last_y] && last_stat==CONFLICTMARK && 
	      last_draw ){
      // we need to undo marked conflict places even 
	// if they are supposed to be empty...
	graph->fCircle(xstart+last_x*width - stonerad, 
	     ystart+width*last_y-stonerad, 2*stonerad, true, true);

  if(last_x!=0 && last_x!=size-1)
    graph->line( xstart+last_x*width-stonerad, ystart+last_y*width,
	      xstart+last_x*width+stonerad, ystart+last_y*width);
  else if (last_x==0){
    graph->line( xstart+last_x*width, ystart+last_y*width,
	      xstart+last_x*width+stonerad, ystart+last_y*width);
    if(last_y!=0 && last_y!=size-1) graph->line( xstart -1, 
		       ystart+last_y*width-stonerad, xstart-1, ystart+last_y*width+stonerad);
    else if (last_y==0)
      graph->line( xstart -1, ystart-1, 
		xstart-1, ystart+last_y*width+stonerad);
    else 
      graph->line( xstart -1, ystart+last_y*width-stonerad, 
		xstart-1, ystart+last_y*width+1);

  }
  else{ // x=size-1
    graph->line( xstart+last_x*width-stonerad, ystart+last_y*width,
	      xstart+last_x*width, ystart+last_y*width);
    if(last_y!=0 && last_y!=size-1) 
      graph->line( xstart +last_x*width+1, 
		ystart+last_y*width-stonerad, xstart+last_x*width+1, 
		ystart+last_y*width+stonerad);
    else if (last_y==0)
      graph->line( xstart +last_x*width+1, ystart-1, 
		xstart+last_x*width+1, ystart+last_y*width+stonerad);
    else 
      graph->line( xstart +last_x*width+1, 
		ystart+last_y*width-stonerad, 
		xstart+last_x*width+1, ystart+last_y*width+1);
  
  }

  if(last_y!=0 && last_y!=size-1)
    graph->line( xstart+last_x*width, ystart+last_y*width-stonerad,
	      xstart+last_x*width, ystart+last_y*width+stonerad);
  else if (last_y==0){
    graph->line( xstart+last_x*width, ystart+last_y*width,
	      xstart+last_x*width, ystart+last_y*width+stonerad);
    if(last_x!=0 && last_x!=size-1)
      graph->line(xstart+last_x*width-stonerad,ystart-1,
		xstart+last_x*width+stonerad, ystart-1);
    else if (last_x==0)
      graph->line(xstart-1,ystart-1,
		xstart+last_x*width+stonerad, ystart-1);
    else
      graph->line(xstart+last_x*width-stonerad,ystart-1,
		xstart+last_x*width+1, ystart-1);


  }
  else { // y=size-1
    graph->line( xstart+last_x*width, ystart+last_y*width-stonerad,
	      xstart+last_x*width, ystart+last_y*width);

    if(last_x!=0 && last_x!=size-1)
      graph->line(xstart+last_x*width-stonerad,ystart+last_y*width+1,
		xstart+last_x*width+stonerad, ystart+last_y*width+1);
    else if (last_x==0)
      graph->line(xstart-1,ystart+last_y*width+1,
		xstart+last_x*width+stonerad, ystart+last_y*width+1);
    else
      graph->line(xstart+last_x*width-stonerad,ystart+last_y*width+1,
		xstart+last_x*width+1, ystart+last_y*width+1);

  }

  if(isStar[last_x][last_y]) markStar(last_x,last_y);
  /*  if(x!=0 && isOccupied[x-1][last_y]) drawStone(x-1,last_y,isWhite[x-1][last_y]);
  if(last_y!=0 && isOccupied[last_x][last_y-1]) drawStone(last_x,last_y-1,isWhite[last_x][last_y-1]);
  if(last_x!=size-1 && isOccupied[last_x+1][last_y]) drawStone(last_x+1,last_y,isWhite[last_x+1][last_y]);
  if(last_y!=size-1 && isOccupied[last_x][last_y+1]) drawStone(last_x,last_y+1,isWhite[last_x][last_y+1]);
  */
      }
	
    last_x=x;
    last_y=y;
    last_stat=stat;
  }
  if(nowQuiet){
    last_draw = false;
    return;
  }

  last_draw = true;

  if(last_stat==REMOVEMARK) return;
  // actually, this here is a major invasion of Boxes' privacy in the upper
  // part of the board.
  if(last_stat == PASSMARKB){
    graph->dString(windowwidth-2*FONTWIDTH-1, 4*BUTTONSIZE-3,"BP");
    return;
  }
  if(last_stat == PASSMARKW){
    graph->dString( windowwidth-2*FONTWIDTH-1,
		    4*BUTTONSIZE-3,"WP");
    return;
  }

  if(last_x >= 0 && isOccupied[last_x][last_y]){
    bool black = (isWhite[last_x][last_y])?BLACK:WHITE;
    if(last_stat == PLUS){
      graph->line(xstart+last_x*width-1,ystart+last_y*width, 
		  xstart+last_x*width+1,ystart+last_y*width, black);
      graph->line(xstart+last_x*width,ystart+last_y*width-1, 
		  xstart+last_x*width,ystart+last_y*width+1, black);
    }
    else if (last_stat==DOT)
      graph->point(xstart+last_x*width, ystart+last_y*width, black);
    else if(last_stat == TIMES){
      graph->line(xstart+last_x*width-1,ystart+last_y*width-1, 
		  xstart+last_x*width+1,ystart+last_y*width+1, black);
      graph->line(xstart+last_x*width+1,ystart+last_y*width-1, 
		  xstart+last_x*width-1,ystart+last_y*width+1, black);

    }
    else if(last_stat == FOURDOTS){
      graph->point( xstart+width*last_x-1, ystart+width*last_y-1, black);
      graph->point( xstart+width*last_x+1, ystart+width*last_y-1, black);
      graph->point( xstart+width*last_x-1, ystart+width*last_y+1, black);
      graph->point( xstart+width*last_x+1, ystart+width*last_y+1, black);
    }
    // for a ko:
    else 
      graph->rectangle( xstart+last_x*width -1, ystart+last_y*width-1, 2,2,
			black);
  }
  else if(last_stat==CONFLICTMARK && !isOccupied[last_x][last_y]){ 
    graph->rectangle( xstart+last_x*width -2, ystart+last_y*width-2, 4,4);
    graph->backRectangle (xstart+last_x*width -1, ystart+last_y*width-1, 3,3);
  }

}

void GoBoard::clear(int which){
  if(which ==1 || which == 0)
    for(int i=0; i<size; i++)
      for(int j=0; j<size; j++)
	ch1[i][j] = false;
  if(which ==2 || which == 0)
    for(int i=0; i<size; i++)
      for(int j=0; j<size; j++)
	ch2[i][j] = false;
  if(which == 3 || which == 0)
    for(int i=0; i<size; i++)
      for(int j=0; j<size; j++)
	ch3[i][j] = false;
  if(which == 4 || which == 0)
    for(int i=0; i<size; i++)
      for(int j=0; j<size; j++)
	ch4[i][j] = false;
}

void GoBoard::markO(int x,int y){
   ch1[x][y]=true;
  if(x!=0 && !ch1[x-1][y]&&!isOccupied[x-1][y])
    markO(x-1,y);
  if(x!=size-1 && !ch1[x+1][y]&&!isOccupied[x+1][y]) 
    markO(x+1,y);
  if(y!=0 && !ch1[x][y-1]&&!isOccupied[x][y-1])
    markO(x,y-1);
  if(y!=size-1 && !ch1[x][y+1]&&!isOccupied[x][y+1]) 
    markO(x,y+1);
}

int GoBoard::fill (int x, int y, bool White){
  if(isOccupied[x][y]) return false;
  clear(1);
  markO(x,y);
  int howmany=0;
  for(int i=0; i<size; i++)
    for(int j=0; j<size; j++)
      if(ch1[i][j]){
	howmany++;
	isOccupied[i][j]=true;
	isWhite[i][j]=White;
	drawStone(i,j,White);
    }
  return howmany;
}

bool GoBoard::koAlone(int x, int y){
  int lib = 0, friends=0;
  if(x!=0)
    if(!isOccupied[x-1][y]) lib++;
    else if (isWhite[x-1][y] == isWhite[x][y]) friends++;
  if(x!=size-1)
    if(!isOccupied[x+1][y]) lib++;
    else if (isWhite[x+1][y] == isWhite[x][y]) friends++;
  if(y!=0)
    if(!isOccupied[x][y-1]) lib++;
    else if (isWhite[x][y-1] == isWhite[x][y]) friends++;
  if(y!=size-1)
    if(!isOccupied[x][y+1]) lib++;
    else if (isWhite[x][y+1] == isWhite[x][y]) friends++;

  if(friends == 0 && lib == 1) return true;
  return false;
}

void GoBoard::initHandi(int xval[], int yval[], int number, bool color, bool quiet){
  for (int i=0; i < size; i++)
    for(int j=0; j<size; j++)
      isOccupied[i][j] = false;
  for(int i=0; i<number; i++){
    isOccupied[xval[i]][yval[i]] = true;
    isWhite[xval[i]][yval[i]] = color;
  }
  if(!quiet) draw();
}

void GoBoard::drawHalfStone(int x, int y, int White){
  if(White == NO_POINT) return;
  if(White == WH){
    graph->fCircle(xstart+x*width - stonerad/2, ystart+width*y-stonerad/2, 
		   2*(stonerad/2), WHITE);
  }
  else {
    graph->fCircle( xstart+x*width - stonerad/2, ystart+width*y-stonerad/2, 
		    2*(stonerad/2), BLACK);
  }
  graph->circle( xstart+x*width - stonerad/2, ystart+width*y-stonerad/2, 
		 2*(stonerad/2), BLACK);
}



//mark_bound marks the boundary AROUND x,y in ch2
// and returns the number of points contained in that territory
int GoBoard::markBound(int x, int y){
  int ret = 1;
  ch1[x][y] = true;
  if(x>0) 
    if(isOccupied[x-1][y] && 
       terr[x-1][y] == NOBODY) 
      ch2[x-1][y] = true;
    else if(!ch1[x-1][y]) ret+= markBound(x-1,y);
  if(y>0)
    if(isOccupied[x][y-1] && 
       terr[x][y-1]  == NOBODY) 
      ch2[x][y-1] = true;
    else if(!ch1[x][y-1]) ret+= markBound(x,y-1);
  if(x<size-1)
    if(isOccupied[x+1][y] && 
       terr[x+1][y] == NOBODY) 
      ch2[x+1][y] = true;
    else if(!ch1[x+1][y]) ret+= markBound(x+1,y);
  if(y<size-1)
    if(isOccupied[x][y+1] && 
       terr[x][y+1]  == NOBODY) 
      ch2[x][y+1] = true;
    else if(!ch1[x][y+1]) ret+= markBound(x,y+1);
  return ret;
}

int GoBoard::belongsWho(){
  int i,j;
  bool whose;
  bool found = false;
  for(i=0; i<size; i++)
    for(j=0; j<size; j++)
      if(ch2[i][j]) 
	if(!found) {found = true; whose = isWhite[i][j];}
	else if(isWhite[i][j] != whose) return NOBODY;
  return (whose == WHITE) ? WH : BL;

}

void GoBoard::markTerr(int x, int y, int whose){
  terr[x][y] = whose;
  //  draw_half_stone(x,y,whose);
  if(x>0 && ch1[x-1][y] && terr[x-1][y] != whose) 
    markTerr(x-1,y,whose);
  if(y>0 && ch1[x][y-1] && terr[x][y-1] != whose) 
    markTerr(x,y-1, whose);
  if(x<size-1 && ch1[x+1][y] && terr[x+1][y] != whose) 
    markTerr(x+1,y,whose);
  if(y<size-1 && ch1[x][y+1] && terr[x][y+1] != whose) 
    markTerr(x, y+1, whose);
}

bool GoBoard::isEye (int x, int y){
      bool copych2[size][size];
      bool copych1[size][size];
      
      for(int i=0; i<size; i++)
	for(int j=0; j<size; j++){
	  copych1[i][j] = ch1[i][j];
	  copych2[i][j] = ch2[i][j];
	}
      markBound(x,y);
      int who = belongsWho();
      for(int i=0; i<size; i++)
	for(int j=0; j<size; j++){
	  ch1[i][j] = copych1[i][j];
	  ch2[i][j] = copych2[i][j];
	}
      return (who!= NOBODY);
}

//counts the inside liberties; recognises "outside inside" spots
int GoBoard::countInsideLibs(int x, int y, bool White){
  int ret = 0, who;
  ch3[x][y] = true;

  if(x>0) 
    if (((isOccupied[x-1][y]&&isWhite[x-1][y] == White) 
	 || terr[x-1][y] == NO_POINT) &&!ch3[x-1][y]) 
      ret+=countInsideLibs(x-1,y, White);
    else if(terr[x-1][y] != NOBODY && !ch3[x-1][y]) {
      ch3[x-1][y] = true;
      ret++;
    }

  if(y>0) 
    if (((isOccupied[x][y-1] && isWhite[x][y-1] == White) 
	 || terr[x][y-1] == NO_POINT) && !ch3[x][y-1]) 
      ret+=countInsideLibs(x,y-1, White);
    else if(terr[x][y-1] != NOBODY && !ch3[x][y-1]) {
      ch3[x][y-1] = true;
      ret++;
    }


  if(y<size-1) 
    if (((isOccupied[x][y+1] && isWhite[x][y+1] == White) 
	 || terr[x][y+1] == NO_POINT) && !ch3[x][y+1]) {
      ret+=countInsideLibs(x,y+1, White);}
    else if(terr[x][y+1] != NOBODY && !ch3[x][y+1]) {
      ch3[x][y+1] = true;
      ret++;
    }


  if(x<size-1) 
    if (((isOccupied[x+1][y] && isWhite[x+1][y] == White)
	 || terr[x+1][y] == NO_POINT) && !ch3[x+1][y]) 
      ret+=countInsideLibs(x+1,y, White);
    else if(terr[x+1][y] != NOBODY && !ch3[x+1][y]) {
      ch3[x+1][y] = true;
      ret++;
    }

  return ret;
}

void GoBoard::makeNeutral(int x, int y){
  ch4[x][y] = true;

  if(x>0) 
    if ((ch2[x-1][y] || terr[x-1][y] == NO_POINT)&&!ch4[x-1][y]) 
      makeNeutral(x-1,y);
    else if(terr[x-1][y] != NOBODY && !ch4[x-1][y]) {
      ch4[x-1][y] = true;
      terr[x-1][y] = NO_POINT;
    }

  if(y>0) 
    if ((ch2[x][y-1] || terr[x][y-1] == NO_POINT) && !ch4[x][y-1]) 
      makeNeutral(x,y-1);
    else if(terr[x][y-1] != NOBODY && !ch4[x][y-1]) {
      ch4[x][y-1] = true;
      terr[x][y-1] = NO_POINT;
    }

  if(y<size-1) 
    if ((ch2[x][y+1] || terr[x][y+1] == NO_POINT) && !ch4[x][y+1]) 
      makeNeutral(x,y+1);
    else if(terr[x][y+1] != NOBODY && !ch4[x][y+1]) {
      ch4[x][y+1] = true;
      terr[x][y+1] = NO_POINT;
    }

  if(x<size-1) 
    if ((ch2[x+1][y] || terr[x+1][y] == NO_POINT) && !ch4[x+1][y]) 
      makeNeutral(x+1,y);
    else if(terr[x+1][y] != NOBODY && !ch4[x+1][y]) {
      ch4[x+1][y] = true;
      terr[x+1][y] = NO_POINT;
    }

}

void GoBoard::computeNeutrals(){
  int i,j;
  clear(3);
  for(i=0; i<size; i++)
    for(j=0;j<size; j++)
      if(ch2[i][j] && !ch3[i][j]){    // new boundary point?
	clear(3);
	if(countInsideLibs(i,j,isWhite[i][j]) < 2) {
	  clear(4);
	  makeNeutral(i,j);
	}
      }
    
}

bool GoBoard::tryChange(int& x, int& y){
  bool hasChanged = false;
  bool changed = true;
  bool stop=false;
  int i,j;
  int oldi=-1, oldj=-1, oldii=-1, oldjj=-1;
  while(changed && !stop){
    changed = false;
    for(i=0; i<size; i++)
      for(j=0; j<size; j++)
	if(terr[i][j] == NO_POINT){
	  //	  fprintf(stderr, "%d, %d\n",i,j);
	  int what = changecount(i,j);
	  //	  printf("%d,%d, %d, %d\n", i, j, terr[i][j], terr[8][0]);
	  changecount(i,j, what);
	  //	  printf("%d,%d, %d, %d\n", i, j, terr[i][j], terr[8][0]);
	  if(terr[i][j]!= NO_POINT) {
	    //	    printf("%d%c,%d,%d\n", i, j+'A', oldi, oldj);
	    //	    drawStone(i, j, BLACK);
	    changed = true;
	    // are we in a loop?
	    if(i==oldii && i==oldi && j==oldjj && j==oldj) stop=true; 
	    hasChanged = true;
	    x=i; y=j; oldii=oldi; oldjj=oldj; oldi=i; oldj=j;
	  }
	}
  }
  //  if(stop==true) printf("TRUE\n");
  return hasChanged;
}

void GoBoard::doChanges(){
  int i,j;
  int x,y,x2,y2;
  int oldx=-1, oldy=-1;
  bool stop=false;

  while(!stop && tryChange(x,y)){
    if(x==oldx && y == oldy) {stop=true; break;}
    oldx=x; oldy=y;
    for(i=0; i<size; i++)
      for(j=0; j<size; j++)
	if(terr[i][j] == NO_POINT){
	  clear(1);
	  clear(2);
	  markBound(i,j);
	  /*
	  for(int l = 0; l<size; l++){
	    fprintf(stderr, "\n");
	    for(int m = 0; m<size; m++)
	      fprintf(stderr, "%c", ((ch2[m][l])?'b':'e'));
	  }
	  fprintf(stderr, "\n");
  */
	  int who = belongsWho();
	  terr[i][j] = ((who == WH)? WH : BL);
	  if(!tryChange(x,y)) terr[i][j] = NO_POINT;
	  else {
	    terr[i][j] = NO_POINT;
	    if(!tryChange(x2,y2)) terr[x][y] = NO_POINT;
	  }
	}
  }
}


void GoBoard::initialCount(int& black, int& white){
  int i,j, whose, number;
  clear(0);
  for(i=0; i<size; i++)
    for(j=0; j<size; j++){
      terr[i][j] = NOBODY;
    }
  
  for(i=0; i<size; i++)
    for(j=0; j<size; j++){
      if(!(ch1[i][j] || isOccupied[i][j])){
	clear(2);
	number = markBound(i,j);
	whose = belongsWho();
	if(whose!= NOBODY){
	  markTerr(i,j,whose);
	  computeNeutrals();

	}
      }
    }

  doChanges();
  
  for(i=0; i<size; i++)
    for(j=0; j<size; j++){
      if(terr[i][j] == WH) white++;
      else if(terr[i][j] == BL) black++;
      if(terr[i][j] != NOBODY && isOccupied[i][j])
	if(isWhite[i][j] == WHITE) black++;
	else white++;
    }
  drawTerritory();
}

void GoBoard::resetTerr(){
  int i,j;
  for(i=0; i<size; i++)
    for(j=0; j<size; j++)
      terr[i][j] = NOBODY;
}

bool GoBoard::performCountAction(int x, int y, int& black, int& white){
  bool retval = false;
  if(isOccupied[x][y] && terr[x][y] == NOBODY){
    countremove(x,y,black,white);
    retval = true;
  }

  doChanges();
  for(int i=0; i<size; i++)
    for(int j=0; j<size; j++){
      if(terr[i][j] == WH) white++;
      else if(terr[i][j] == BL) black++;
      if(terr[i][j] != NOBODY && isOccupied[i][j])
	if(isWhite[i][j] == WHITE) black++;
	else white++;
    }
  
  drawTerritory();
  return retval;
}

int GoBoard::changecount(int x, int y, int whose){
  bool done = false;
  bool compare = (whose == WH) ? WHITE: BLACK;

  int i,j;
  clear(1);
  clear(2);
  markBound(x,y);
  
  if(whose != NOBODY){
  
    i=x; j=y;
    while(!done){
      done = true;
      clear(1);
      clear(2);
      markBound(i,j);
      for(i=0; i<size && done; i++){
       for(j=0; j<size && done; j++){
	if(ch2[i][j] && (isWhite[i][j] != compare || !isOccupied[i][j]))
	  {
	  done = false;
	  terr[i][j] = SET_LATER;
	}
      }
    }
    i--;
    j--;
  }

  }
  /*
  fprintf(stderr, "\nstart\n");
    for(int i=0; i<size; i++){
      for(int j=0; j<size; j++){
	if(!ch1[j][i]) fprintf(stderr, "%c",((ch2[j][i])?'b':'e'));
	else fprintf(stderr, "t");
      }
      fprintf(stderr, "\n");
    }
  */
  if((whose=belongsWho()) == NOBODY){fprintf(stderr, "RET\n"); return whose;}
  if(terr[x][y] != NOBODY && terr[x][y] != SET_LATER){
    for(int i=0; i<size; i++){
      for(int j=0; j<size; j++){
	if(ch1[i][j]) terr[i][j] = NOBODY;
      }
    }
  }
  else{
    markTerr(x,y,whose);
    computeNeutrals();
  }
  return whose;
  
}

void GoBoard::countremove(int x, int y, int& black, int& white){
  int i,j;
  bool done = false;
  int toWhom = WH;
  if(isWhite[x][y] == WHITE) toWhom = BL;
  terr[x][y] = SET_LATER;

  i=x; j=y;
  while(!done){
    done = true;
    clear(1);
    clear(2);
    markBound(i,j);
     for(i=0; i<size && done; i++){
       for(j=0; j<size && done; j++){
	if(ch2[i][j] && (isWhite[i][j] == isWhite[x][y] || !isOccupied[i][j]))
	  {
	  done = false;
	  terr[i][j] = SET_LATER;
	}
      }
    }
    i--;
    j--;
  }
  markTerr(x,y,toWhom);
  clear(3);
  computeNeutrals();
}

bool GoBoard::isOcc(int x, int y){
  if(x<0 || y<0 || x>= size || y >= size) return false;
  return isOccupied[x][y];
}
