/*
GoControl is the object which handles most of what is actually going on.
govrec.cc is just the interface.
 */

class GoControl{
private:
  Move* current;
  Move* zero;
  GoBoxes* TBoxes;
  GoBoard* Board;
  GoGraphic* graph;
  //  MSB* selMoves;
  GoFile* File;
  bool coordinates;
  int obstruction, s_height, s_width;
  int width, xstart, ystart;
  int boardsize;
  int handicp;
  int handix[9];
  int handiy[9];
  int mr, black, white;
  int moveNumber;
  int lastX, lastY;
  
  int komi;
  int handicap;
  bool whiteplusdotfive;
  bool conflict;
  int replace;
  bool defaultSwitch;
  bool changed;
  bool autosave, sound, gamestart;

  void readFile();
  bool performNextMove (bool quiet = false);
  void removeLastStone(bool quiet = false);

  void initBoard();
  bool doMove(int i, int j, int what = OK);
  void handleBoxButton (int w);
  void handleBoardClick(int,int);
  void goBackTo(int,int);
  void initVariables();
  void initHandicap();
  void rotateHandi();
  void count();
  void playsound(int);
  void quit();
  void drawQuit();
  void drawOverwrite(const char* filen);
  //  void resetKeyboard(int,void*);
#ifndef USE_QT
  GC CreateGC(Display *display, Drawable drawable, unsigned long forecolor,
	      unsigned long backcolor);
#endif

public:
  GoControl(int nargs, char** args, GoGraphic*);
  void resize();
  void handleGameClick(int, int);
  void handleOtherClick(int);
  void handleInitClick(int,int);
  void handleKeyPress(GoEvent&);

  void exposed();
  void initialiseNewGame();
  void gameStart();
  bool gameStarted();
  void autoSave();

};

GoControl::GoControl(int nargs, char** args, GoGraphic* gr){
  autosave = ALWAYSAUTOSAVE;
  TBoxes = 0;
  boardsize = DEFAULTBOARDSIZE;
  coordinates = DEFAULT_COORDINATES;
  File = new GoFile((nargs == 2)? args[1]: "");
  graph = gr;
  char filename[MAXSTRING+1];
  graph->setupWindow(File->getFilename(filename), nargs, args);
  graph->setSizes(&s_width, &s_height);
  obstruction = ((s_width)>s_height) ? (s_width) : s_height;
  initVariables();
  komi=5;
  Board    = new GoBoard(graph, s_width, s_height);
  //  selMoves = new MSB(graph);
  //  Boxes = new TopBoxes(graph);

  if (nargs == 2) readFile();        // parse the input file
}

void GoControl::gameStart(){
  //  komi = white;
  gamestart = true;
  if(handicap != 0) zero->White = BLACK;

  graph->emptyRectangle(0,0,windowwidth-BUTTONSIZE+1,BUTTONSIZE);

  TBoxes->setGameStart();
  
  TBoxes->setBoardProp(boardsize, xstart, width);
  TBoxes->redraw(current);
  //  selMoves->drawChoices(current);

  File->setGameProp(boardsize, handicap, komi, whiteplusdotfive);
  //  fprintf(stderr, "gameStartK %d\n",komi);
}

void GoControl::initVariables(){
  gamestart = false;
  current = zero = 0;
  lastX = lastY = 0;
  black = white = moveNumber = 0;
  replace = MOVEBUTTON;
  defaultSwitch=true;
  changed = false;
  whiteplusdotfive=KOMI_DOT_FIVE;
  conflict=false;
  boardsize = 19;
  sound = SOUNDDEFAULT;

  handicap=0;
  //  Board = 0;
  //  selMoves = 0;
  delete TBoxes;
  TBoxes = new GoBoxes(boardsize, 0, whiteplusdotfive, graph);
}

void GoControl::initialiseNewGame(){
  initVariables();
  zero = new Move(-1,-1,WHITE, OK);
  current = zero;
  white = KOMI_INTEGER;
  komi=white;

  initBoard();
  Board->redraw();
  TBoxes->setKomi(white, whiteplusdotfive);
  TBoxes->setBoardProp(boardsize, xstart, width);
  TBoxes->draw(moveNumber,black,white, replace, current);
  TBoxes->redraw(current);
}

void GoControl::handleInitClick(int x, int y){
  int whichButton=TBoxes->isInButton(x,y);
  if(sound) if (whichButton==NOTHING) graph->playsound (ALARMTONE);
  else graph->playsound(SYSTEMTONE);
  switch(whichButton){                      // introbuttons
  case DECR_SIZE:                           // left size
    if(boardsize == 19) {
      boardsize = 13;
      white = KOMI_INTEGER;
    }
    else if (boardsize == 13) {
      boardsize = 9;
      white = 0;
    }
    else {
      boardsize = 19;
      white = KOMI_INTEGER;
    }
    whiteplusdotfive = KOMI_DOT_FIVE;
    defaultSwitch = true;
    handicap = 0;
    initBoard();
    break;
  case INCR_SIZE:                           // right size
    if(boardsize == 19) {
      boardsize = 9;
      white = 0;
    }
    else if (boardsize == 13) {
      boardsize = 19;
      white = KOMI_INTEGER;
    }
    else {
      boardsize = 13;
      white = KOMI_INTEGER;
    }
    whiteplusdotfive = KOMI_DOT_FIVE;
    defaultSwitch = true;
    handicap = 0;
    initBoard();
    break;
  case DECR_HANDI:                        // left handi
    handicap --;
    if(handicap == 1) {
      handicap = 0;
      if(boardsize==9) white = 0;
      else white = KOMI_INTEGER;
      whiteplusdotfive = KOMI_DOT_FIVE;
      defaultSwitch = true;
    }
    else if(handicap <= 0) {
      handicap = 9;
      white = 0;
      whiteplusdotfive = KOMI_DOT_FIVE;
      defaultSwitch = true;
    }
    TBoxes->setHandi(handicap);
    initHandicap();
    break;
  case INCR_HANDI:                      // right handi
    handicap = (handicap+1)%10;
    if(handicap == 1) {
      handicap = 2;
      white = 0;
      whiteplusdotfive = KOMI_DOT_FIVE;
      defaultSwitch = true;
    }
    if(handicap == 0) {
      white = KOMI_INTEGER;
      whiteplusdotfive = KOMI_DOT_FIVE;
      defaultSwitch = true;
    }
    TBoxes->setHandi(handicap);
    initHandicap();
    break;
  case ROT_HANDI:                      // rotate handi
    rotateHandi();
    break;
  case DECR_KOMI:                      // decrease komi
    if(defaultSwitch && white == KOMI_INTEGER){
      white = 0;
      break;
    }
    if(whiteplusdotfive == !KOMI_DOT_FIVE && 
       ((!KOMI_DOT_FIVE && white == KOMI_INTEGER) || 
	          (KOMI_DOT_FIVE && white == KOMI_INTEGER+1))){
      white = KOMI_INTEGER;
      whiteplusdotfive = KOMI_DOT_FIVE;
      defaultSwitch = true;
      break;
    }
    if(defaultSwitch && white == 0){
      white = KOMI_INTEGER;
      whiteplusdotfive = !KOMI_DOT_FIVE;
      if( !KOMI_DOT_FIVE ) white--;
      defaultSwitch = false;
      break;
    }
    if(whiteplusdotfive){
      whiteplusdotfive = false;
    }
    else{
      white --;
      if(white < 0) white = 9;
      whiteplusdotfive = true;
    }
    defaultSwitch = false;
    break;
  case INCR_KOMI:                     // increase komi
    if(defaultSwitch && white == 0){
      white = KOMI_INTEGER;
      break;
    }
    if(whiteplusdotfive == !KOMI_DOT_FIVE && 
       ((KOMI_DOT_FIVE && white == KOMI_INTEGER) || 
	          (!KOMI_DOT_FIVE && white == KOMI_INTEGER-1))){
      white = 0;
      whiteplusdotfive = KOMI_DOT_FIVE;
      defaultSwitch = true;
      break;
    }
    if(!whiteplusdotfive)
      whiteplusdotfive = true;
    else{
      white = (white +1)%10;
      whiteplusdotfive = false;
    }
    defaultSwitch = false;
    break;
  }
  Board->redraw();
  TBoxes->setBoardProp(boardsize, xstart, width);
  komi=white;
  TBoxes->setHandi(handicap);
  TBoxes->setKomi(white, whiteplusdotfive);
  TBoxes->draw(moveNumber, black, white, replace, current);
  TBoxes->redraw(current);
  
}

void GoControl::handleGameClick(int x, int y){
  int whichButton = TBoxes->isInButton(x, y);
  //  if(whichButton != NOTHING){
  //    handleBoxButton(whichButton);
  //  }
  //  else{
    //    whichButton = selMoves->isInButton(x, y);
    if(whichButton != NOTHING){
      if(whichButton>=0) {
	current->nextNumber = whichButton;
	performNextMove ();
      }
      else handleBoxButton(whichButton);
      if(sound) graph->playsound(SYSTEMTONE);
    }
    else{
      if(y < ystart-width/3) {
	lastX = -1;
	if(sound) graph->playsound(ALARMTONE);
      }
      else{
	int i,j;
	i = (int)((((float) x)-xstart)/width + .5);
	j = (int)((((float) y)-ystart)/width + .5);
	handleBoardClick(i,j);
    }
    }
  TBoxes->draw(moveNumber, black, white, replace, current);
//selMoves->drawChoices(current);
}  

void GoControl::handleBoardClick(int i, int j){           //reg.Move
  if(0 > i || i >= boardsize || j < 0 || j >= boardsize){
    lastX = -1;
    if(sound) graph->playsound(ALARMTONE);
    return;
  }

  if(Board->isOcc(i,j)) {
    if(sound) graph->playsound(SYSTEMTONE);
    if(i == lastX && j == lastY){
      goBackTo(i,j);
      lastX=-1; 
      return;
    }
    else {
      lastX = i;
      lastY = j;
      return;
    }
  }
  if(doMove(i,j)){
    if (sound) graph->playsound(STONETONE);
  }
  else if (sound) graph->playsound(ALARMTONE);
  lastX = -1;
}

bool GoControl::doMove(int i, int j, int what){
  bool retVal = true;
  if(replace != REPLACEBUTTON){
    if(what == PASS && (current->what & PASS) != 0) {
      count(); 
      
      if(current->hasNext()){
	replace = DEFAULTREPLACEVALUE;
      }
      else{
	replace = MOVEBUTTON;
      }
      return true;
    }

    if(what != PASS){
      // regular move or variation?
      what=(Board->setStone(i,j,!current->White, 
			    (!current->White)?white:black));
    }
    else Board->markStone(0,0,
			   (!current->White == WHITE)?PASSMARKW : PASSMARKB);

    if((what&OK)==OK || what == PASS){
      changed = true;
      if(replace != ADDBUTTON){
	current= current->nextAppend(i,j,!current->White, what);
	// append OR find the variation which features this move
	//		current = current->getNext(current->nextNumber);
	TBoxes->draw(moveNumber, black, white, replace, current);
	replace=MOVEBUTTON;
      }
      else
	current = current ->insertBeforeNext(i,j,!current->White, what);

      moveNumber++;
      
    }
    else 
      retVal = false;
  }
  else{                                     //replace Stone
    if(!conflict && (current->what)!= PASS)
      Board->removeStone(current->x, current->y, current->what,
			 (current->White)?white:black);
    
    if(what != PASS)
      what =(Board->setStone(i,j,current->White,(current->White)?white:black));
    else Board->markStone (0,0,(current->White)?PASSMARKW:PASSMARKB);
    
    if((what & OK) == OK || what == PASS){
      changed = true;
      current->x = i;
      current->y = j;
      current->what = what;
      
      conflict=false;
      TBoxes->setConflict(false);
#if defined REPLACE_WILL_ADVANCE
      performNextMove ();
#else
      if(current->hasNext()){
	replace = DEFAULTREPLACEVALUE;
      }
      else{
	replace = MOVEBUTTON;
      }
#endif
    }
    else {
      if(current->what != PASS){
	Board->setStone(current->x, current->y, 
			current->White,(current->White)?white:black);
	retVal = false;
      }
    }
  }
  if(replace!=MOVEBUTTON && replace != VARIATIONBUTTON && !conflict)
    Board->markStone(current->x, current->y, (replace==REPLACEBUTTON)?
		     REPLACEMARK : ADDMARK);
  lastX = -1;
  return retVal;
}

void GoControl::goBackTo(int i, int j){
  while((current->x != i || current->y != j || current->what == PASS || conflict) 
	&& current != zero){
    removeLastStone(true);
  }
  if(current != zero){
    removeLastStone(true);
    performNextMove (true);
  }
  if(current->hasNext()) replace = DEFAULTREPLACEVALUE;
  Board->redraw();
}

void GoControl::handleBoxButton (int w){
  switch (w){
  case SCROLLDOWN:
    TBoxes->draw(moveNumber, black, white, replace, current, COM_SCROLLDOWN);
    break;
  case SCROLLUP:
    TBoxes->draw(moveNumber, black, white, replace, current, COM_SCROLLUP);
    break;
  case TOGGLEAUTOSAVE:
    autosave = !autosave;
    break;
  case TOGGLECOORD:
    coordinates = !coordinates;
    resize();
    break;
  case TOGGLESOUND:
    sound = !sound;
    break;
  case QUIT:
    if(sound) graph->playsound(SYSTEMTONE);
    quit();
    break;
  case MVR:
    if(sound) graph->playsound(SYSTEMTONE);
    if (current == 0 || !(current -> hasNext())){
      if(moveNumber!=0){
	if(replace == REPLACEBUTTON){
	  replace = MOVEBUTTON;
	  if((current->what)!=PASS)
	    Board->markStone(current->x, current->y, NORMALMARK);
	}
	else{
	  replace = REPLACEBUTTON;
	  if((current->what)!=PASS)
	    Board->markStone(current->x, current->y, REPLACEMARK);
	}
      }
      lastX = -1;
      break;
    }
    if(replace == VARIATIONBUTTON) replace = REPLACEBUTTON;
    else if(replace == REPLACEBUTTON) replace = ADDBUTTON;
    else replace = VARIATIONBUTTON;

    if(moveNumber!=0)
      if((current->what)!=PASS)
	if(replace == ADDBUTTON)
	  Board->markStone(current->x, current->y, ADDMARK);
	else
	  Board->markStone(current->x, current->y, 
			   (replace==REPLACEBUTTON)?REPLACEMARK:NORMALMARK);
    break;

  case PASSB:
    if(sound) graph->playsound(SYSTEMTONE);
    doMove(0,0,PASS);
    break;
    
  case TOGAMESTART:
    if (current == zero) {
      if(sound) graph->playsound(ALARMTONE);
      break;
    }
    if(sound) graph->playsound(SYSTEMTONE);
    goBackTo(-1,-1);
    break;
  case DECR_MOVE:
    if(current==zero) {
      if(sound) graph->playsound(ALARMTONE);
      break;
    }
    if(sound) graph->playsound(SYSTEMTONE);
    removeLastStone();
    break;
  case INCR_MOVE:
    if(!current->hasNext()) {
      if(sound) graph->playsound(ALARMTONE);
      break;
    }
    if(sound) graph->playsound(SYSTEMTONE);
    replace = DEFAULTREPLACEVALUE;
    performNextMove ();
      
    break;
  case TOGAMEEND:
    if(!current->hasNext()){
      if(sound) graph->playsound(ALARMTONE);
      break;
    }
    if(sound) graph->playsound(SYSTEMTONE);
    while(current->getNext() != 0&&
      performNextMove (true));
    
    Board->redraw();
    
    TBoxes->draw(moveNumber,black,white, replace, current);
    break;
  }
}

void GoControl::handleOtherClick ( int type){
  if(type == MID_BUTTON)  handleBoxButton(INCR_MOVE);
  else handleBoxButton(DECR_MOVE);

  lastX = -1;
  TBoxes->draw(moveNumber,black,white, replace, current);
  //  selMoves->drawChoices(current);
}

void GoControl::handleKeyPress( GoEvent& event){
  if(((event.xtype) ==PgUP || (event.xtype)==RIGHT)&& current!= 0){ //PgUP
    if(event.shift) handleBoxButton(TOGAMEEND);
    else handleBoxButton(INCR_MOVE);
  }
                                                                 // PgDown
  else if (((event.xtype) ==PgDOWN || (event.xtype)==LEFT) && moveNumber!=0){
    if(event.shift) handleBoxButton(TOGAMESTART);
    else handleBoxButton(DECR_MOVE);
  }
  else if ((event.xtype) ==QUITKEY){
    if(sound) graph->playsound(SYSTEMTONE);
    quit();
  }
  lastX = -1;
  TBoxes->draw(moveNumber,black,white, replace, current);
  //  selMoves->drawChoices(current);
}

void GoControl::autoSave(){
  static int lastmn;
  if(!changed || !autosave) {lastmn = moveNumber; return;}
  if(lastmn == moveNumber) return;
  lastmn = moveNumber;
  File->createFile(zero, true, handix, handiy);
  changed = false;
}


bool GoControl::performNextMove (bool quiet){
  int mr;
  int nextStones;
  if( ( current->hasNext()) && !conflict){
    current=current->getNext();

    if(current->what != PASS){
      mr = Board->setStone(current->x, current->y,
			   (current->White),(current->White)?white:black, 
			   quiet);
      if((mr&OK)==OK)
	current->what = mr;
      else {
	conflict=true;
	TBoxes->setConflict(true);
	replace=REPLACEBUTTON;
	moveNumber++;
	Board->markStone(current->x, current->y, CONFLICTMARK, quiet);
	return !conflict;
      }
      
    }
    else Board->markStone(0,0,(current->White== WHITE)?PASSMARKW : PASSMARKB,
			   quiet); // in case of passing move
    moveNumber++;
    if(!current->hasNext()){
      if(replace != REPLACEBUTTON) replace = MOVEBUTTON;
      else
	if(!quiet)
	Board -> markStone(current->x, current->y, REPLACEMARK);
    }
    else {
      if(!quiet && replace == REPLACEBUTTON)
	Board -> markStone(current->x, current->y, REPLACEMARK);
    }
  }
  else if(!conflict) {
    replace = MOVEBUTTON;
    return !conflict;
  }
  return !conflict;
}

void GoControl::removeLastStone(bool quiet){
  if(!conflict && (current->what != PASS)){
    Board->removeStone((current)->x, (current)->y, (current)->what,
		       (current->White)?white:black, quiet);
  }
  conflict=false;
  TBoxes->setConflict(false);
  moveNumber--;
  
  if((current) != zero) (current) = (current)->prev;
  if(moveNumber != 0) {
    replace = DEFAULTREPLACEVALUE;
    if(((current)->what) == PASS) 
      Board->markStone(0,0,(current->White== WHITE)?PASSMARKW : PASSMARKB);
    else
      if(replace == REPLACEBUTTON) 
	Board -> markStone((current)->x, (current)->y, REPLACEMARK, quiet);
      else
	Board->markStone((current)->x, (current)->y,
			  ((((current)->what)&KO)==0)?NORMALMARK:0, quiet);
  }
  else replace = VARIATIONBUTTON;
}


void GoControl::readFile(){
  // parse the input-file
  zero = File-> parseFile(boardsize, handicap, white, 
			  whiteplusdotfive, handix, handiy);

  if(ALWAYSAUTOSAVEWITHFILENAME) autosave = true;
  if(zero == 0 && AUTOSAVEWITHEMPTYFILE) autosave = true;
  komi = white;
  
  if(zero!= 0){
    if(handicap != 0) zero->White = BLACK;

    initBoard();

    Board->initHandi(handix, handiy, handicap, BLACK, true);
  
    TBoxes->setKomi(white, whiteplusdotfive);
    TBoxes->setBoardProp(boardsize, xstart, width);
    TBoxes->setGameStart();
    TBoxes->redraw(current);
  }
  current = zero;
#if defined loadGameEnd    
  if(zero!=0){
    while(current->getNext() != 0 &&
      performNextMove (true));

  }
#endif
  //  if(current != 0) //selMoves->drawChoices(current);
  if(zero!=0){
    TBoxes->draw(moveNumber,black,white, replace, current);
    Board->redraw();
    gameStart();
  }
}

void GoControl::initBoard(){
  if(!coordinates) {
    width = windowwidth/boardsize;
    xstart = (windowwidth - (boardsize-1)*width)/2;
    ystart = windowheight - (boardsize-1)*width - xstart;
  }
  else {
    width = (windowwidth-obstruction)/boardsize;
    xstart = (windowwidth - (boardsize-1)*width)/2 + s_width/2;
    ystart = (windowwidth - (boardsize-1)*width)/2 - s_height/2 +
      windowheight - windowwidth;
  }

  graph->emptyRectangle(0,0,windowwidth, windowheight);

  //  delete Board;
  Board ->setProperties(boardsize, xstart, ystart, width, coordinates);

  //  delete selMoves;
  TBoxes ->setProperties(width, autosave, sound, coordinates);

  TBoxes->setBoardProp(boardsize, xstart, width);
}

void GoControl::resize(){
  if(!coordinates) {
    width = windowwidth/boardsize;
    xstart = (windowwidth - (boardsize-1)*width)/2;
    ystart = windowheight - (boardsize-1)*width - xstart;
  }
  else {
    width = (windowwidth-obstruction)/boardsize;
    xstart = (windowwidth - (boardsize-1)*width)/2 + s_width/2;
    ystart = (windowwidth - (boardsize-1)*width)/2 - s_height/2 +
      windowheight - windowwidth;
  }

  graph->emptyRectangle(0,0,windowwidth, windowheight);

  //  delete Board;
  Board ->resize( xstart, ystart, width, coordinates);

  //  delete selMoves;
  TBoxes->resize(width, autosave, sound, coordinates, boardsize, xstart, width, current);
  exposed();
}




bool GoControl::gameStarted(){
  return (zero != 0);
}

void GoControl::exposed(){
  graph->emptyRectangle(0,0,windowwidth, windowheight-windowwidth+1);
  //  if(gamestart) selMoves->drawChoices(current);
  Board->redraw();
  TBoxes->redraw(current);
  Board->markStone(0,0,REDRAWMARK);
  //  graph->dSString(10,10,"langerkurzerstring");
  //  graph->doDraw();
}

void GoControl::initHandicap(){
  int count = 0;
  int border = (boardsize > 9) ? 3 : 2;
  if(handicap %2 == 1 && handicap != 3){
    handix[count] = handiy[count] = boardsize/2;
    count++;
  }
  if(handicap != 1){
    handix[count] = boardsize - 1 - border;
    handiy[count] = border;
    count ++;
    handix[count] = border;
    handiy[count] = boardsize - 1 - border;
    count ++;
  }
  if (handicap > 2){
    handix[count] = border;
    handiy[count] = border;
    count ++;
  }
  if(handicap > 3){
    handix[count] = boardsize-1-border;
    handiy[count] = boardsize-1-border;
    count ++;
  }
  if(handicap > 5){
    handix[count] = border;
    handiy[count] = boardsize/2;
    count ++;
    handix[count] = boardsize-1-border;
    handiy[count] = boardsize/2;
    count ++;
  }
  if(handicap>7){
    handiy[count] = border;
    handix[count] = boardsize/2;
    count ++;
    handiy[count] = boardsize-1-border;
    handix[count] = boardsize/2;
    count ++;
  }
  Board->initHandi(handix, handiy, handicap, BLACK);

}

void GoControl::rotateHandi(){
   int interm;
   for(int i=0; i < handicap; i++){
     interm = handix[i];
     handix[i] = boardsize-1-handiy[i];
     handiy[i] = interm;
   }
  graph->backRectangle(0,windowheight-windowwidth,
		 windowwidth, windowwidth);
  Board->initHandi(handix, handiy, handicap, BLACK);
   
 }

void GoControl::count(){
  GoEvent event;
  //  KeySym key;
  int i,j;
  int initBlack = black, initWhite = white;
  bool afterInit = true;
  Board->markStone(0,0,REMOVEMARK);
  Board->initialCount(black, white);

  TBoxes->setCounting();
  
  graph->emptyRectangle(0, 0, windowwidth, 
		 windowheight-windowwidth-FONTHEIGHT);

  TBoxes->draw(moveNumber,black,white, EMPTYBUTTON, current);
  TBoxes->redraw(current);

  while(true){
    graph->getEvent(&event);

    switch(event.type){
    case BUTTON:                             //MouseButton
	if(event.xtype >=windowwidth-BUTTONSIZE)      //right Buttons
	  if( event.y < BUTTONSIZE){
	    if(sound) graph->playsound(SYSTEMTONE);
	    quit();
	    break;
	  }
	i = (int)((((float) event.xtype)-xstart)/width + .5);
	j = (int)((((float) event.y)-ystart)/width + .5);
	if(!(i>=0 && j >= 0 && i < boardsize && j < boardsize)) {
	  if(sound) graph->playsound(ALARMTONE);
	  break;
	}

	black = initBlack;
	white = initWhite;
	if(i>=0 && i<boardsize && j>= 0 && j<boardsize){
	  if(Board->performCountAction(i,j,black,white)){
	    if(sound) graph->playsound(SYSTEMTONE);
	  }
	  else if(sound) graph->playsound(ALARMTONE);

	  afterInit = false;
	}
	TBoxes->draw(moveNumber, black, white, EMPTYBUTTON, current);
	break;
    case KEY:
      if(sound) graph->playsound(SYSTEMTONE);
      
      if ((event.xtype ==PgDOWN||event.xtype==LEFT) && moveNumber!=0){
	if(afterInit) {
	  Board->resetTerr();
	  black = initBlack;
	  white = initWhite;
	  graph->backRectangle(0, windowheight - windowwidth, 
			       windowwidth, windowheight);
	  Board->draw();
	  TBoxes->draw(moveNumber, black, white, replace, current);
	  graph->emptyRectangle(0,0,windowwidth, windowheight-windowwidth);
	  
	  TBoxes->unsetCounting();
	  
	  TBoxes->redraw(current);
	  Board->markStone(0,0,(current->White==WHITE)?PASSMARKW:
			   PASSMARKB);
	  
	  return;
	}
	
	Board->resetTerr();
	graph->backRectangle(0, windowheight - windowwidth, 
			     windowwidth, windowheight);
	black = initBlack;
	white = initWhite;
	Board->draw();
	Board->initialCount(black, white);
	TBoxes->draw(moveNumber, black, white, EMPTYBUTTON, current);
	Board->draw();
	afterInit = true;
      }
      break;
    case RESIZE:
      windowwidth = event.xtype;
      windowheight = event.y;
      if (windowwidth > windowheight - 3*BUTTONSIZE - 2*FONTHEIGHT){
	windowwidth = windowheight - 3*BUTTONSIZE - 2*FONTHEIGHT;
	graph->fRectangle (windowwidth,0,event.xtype - windowwidth, 
			   windowheight, BLACK);
      }
      resize();

    case EXPO:
      exposed();
    }
  }

}

void GoControl::quit(){
  bool cont=true;
  bool askOverwrite = false;
  char filen[MAXSTRING+1];
  if(!changed) exit(0);
  GoEvent event;
  graph->emptyRectangle(0, 0, windowwidth-BUTTONSIZE, 
		 windowheight-windowwidth-FONTHEIGHT); 
  
  drawQuit();
  
  while(cont){
    graph->getEvent(&event);
    
    switch(event.type){
    case BUTTON:                             //MouseButton
      if(sound) graph->playsound(SYSTEMTONE);
      if(event.y >=
	 ((windowheight-windowwidth)/FONTHEIGHT/2)*FONTHEIGHT
	 && event.y <= 
	 ((windowheight-windowwidth)/FONTHEIGHT/2+1)*FONTHEIGHT)
	if(event.xtype >=
	   ((windowwidth)/FONTWIDTH/2-3)*FONTWIDTH
	   && event.xtype <= 
	   ((windowwidth)/FONTWIDTH/2-2)*FONTWIDTH){
	  if(File->createFile(zero, askOverwrite, handix, handiy))
	    exit(0);
	  else {
	    askOverwrite = true;
	    drawOverwrite(File->getFilename(filen));
	  }
	}
	else if(event.xtype >=
		((windowwidth)/FONTWIDTH/2+1)*FONTWIDTH
		&& event.xtype <= 
		((windowwidth)/FONTWIDTH/2+2)*FONTWIDTH)
	  exit(0);
	else
	  cont = false;
      else cont = false;
      break;

    case KEY:
      if(event.xtype == NOKEY){
	  exit(0);
	  break;
      }
      else if(event.xtype == YESKEY){
	  if(File->createFile(zero, askOverwrite, handix, handiy))
	    exit(0);
	  else {
	    askOverwrite = true;
	    drawOverwrite(File->getFilename(filen));
	  }
      }
      break;

    case RESIZE:
      windowwidth = event.xtype;
      windowheight = event.y;
      if (windowwidth > windowheight - 3*BUTTONSIZE - 2*FONTHEIGHT){
	windowwidth = windowheight - 3*BUTTONSIZE - 2*FONTHEIGHT;
	graph->fRectangle (windowwidth,0,event.xtype - windowwidth, 
			   windowheight, BLACK);
      }
      resize();
      if(askOverwrite)
	drawOverwrite(File->getFilename(filen));
      else drawQuit();
      break;

    case EXPO:
      exposed();
      if(askOverwrite)
	drawOverwrite(File->getFilename(filen));
      else drawQuit();
    }
  }

  graph->emptyRectangle(0, 0, windowwidth-BUTTONSIZE-1, 
		 windowheight-windowwidth-FONTHEIGHT); 
  TBoxes->redraw(current);
}

void GoControl::drawQuit(){
  graph->emptyRectangle(0,0, windowwidth, windowheight-windowwidth-FONTHEIGHT);
  graph->dString( ((windowwidth)/(FONTWIDTH*2)-3)*FONTWIDTH,
		  ((windowheight-windowwidth)/(FONTHEIGHT*2)-1)*FONTHEIGHT,
		  "Save?");
  graph->dString( ((windowwidth)/FONTWIDTH/2-3)*FONTWIDTH,
		  ((windowheight-windowwidth)/FONTHEIGHT/2+1)*FONTHEIGHT,
		  "Y C N");
}



void GoControl::drawOverwrite(const char* filen){
  graph->emptyRectangle(0,0, windowwidth, windowheight-windowwidth-FONTHEIGHT);

  graph->dString( (long unsigned int)
		  (((float)windowwidth)/(FONTWIDTH*2)-11.5)*FONTWIDTH,
		  ((windowheight-windowwidth)/(FONTHEIGHT*2)-1)*FONTHEIGHT,
		  "Overwrite existing file");
  graph->dString( ((windowwidth)/(FONTWIDTH*2) - 
		   strlen(filen)/2)*FONTWIDTH,
		  ((windowheight-windowwidth)/(FONTHEIGHT*2))*FONTHEIGHT,
		  filen);
  graph->dString( ((windowwidth)/(FONTWIDTH*2) + 
		   strlen(filen)/2)*FONTWIDTH,
		  ((windowheight-windowwidth)/(FONTHEIGHT*2))*FONTHEIGHT,
		  "?");

  graph->dString( ((windowwidth)/FONTWIDTH/2-3)*FONTWIDTH,
		  ((windowheight-windowwidth)/FONTHEIGHT/2+1)*FONTHEIGHT,
		  "Y C Q");
}

