/*
GoFile takes care of the file operations
Andreas Balser
*/

class GoFile{
private:
  int size;
  int handi;
  int komi;
  bool plusFive;
  char filename[MAXSTRING+1];
  char plWhite[MAXSTRING+1];
  char plBlack[MAXSTRING+1];
  char gameName[MAXSTRING+1];

  
  void gotoNext(FILE* sgf, char ch);
  char gotoEither(FILE* sgf, char ch, char ch2);
  char gotoEither(FILE* sgf, char ch, char ch2, char ch3);
  char fgetchar(FILE* file);
  void parseString(FILE* sgf, char* string);
  Move* getInitialMoves (FILE* sgf, Move* cur, bool white);
  char parseAppend(FILE* sgf, Move* cur);
  void writeFile(FILE* sgf, Move* first);
  void getMoveString(char ** move, Move* first);

  

public:
  GoFile(const char*);
  void setGameProp(int,int,int, bool);
  char* getFilename(char*);
  Move* parseFile(int&, int&, int&, bool&, int*, int*);
  bool createFile(Move*, bool, int*, int*);
};

GoFile::GoFile(const char* fn){
  strcpy(filename, fn);

  strcpy(plWhite, PLAYERWHITE);
  strcpy(plBlack, PLAYERBLACK);
  strcpy(gameName, plWhite);
  strcat(gameName, " (W) vs. ");
  strcat(gameName, plBlack);
  strcat(gameName, "(B)");
}

void GoFile::setGameProp(int s, int h, int k, bool kb){
  size = s;
  handi = h;
  komi = k;
  //  fprintf(stderr, "gamepropK %d\n", komi);
  plusFive = kb;
}

char* GoFile::getFilename(char* fn){
  strcpy(fn, filename);
  return fn;
}

bool GoFile::createFile(Move* zero, bool forceOverwriteAndMakeSafetyCopy, 
			int* handix, int* handiy){
  FILE * sgf;
  time_t now = time (0);
  char stri[20];
  tm* day = localtime (&now);
  bool useExFilename = false;
  char safetycopy[2*MAXSTRING];
  static int MakeSafetyCopy = MAKESAFETYCOPY;
  safetycopy[0]=0;
  if((forceOverwriteAndMakeSafetyCopy && filename[0]!=0) || 
                        (strcmp(filename,FILENAME)!= 0 
			&& strcmp(filename, ALTERNATEFILE)!= 0
			&& strcmp(filename, THIRDFILE) != 0)){
    if((sgf=fopen(filename,"w"))!=0){
      useExFilename = true;
      strcpy (safetycopy, COPYCOMMAND);
      strcat(safetycopy, " ");
      strcat(safetycopy, filename);
      strcat(safetycopy, " ");
      strcat(safetycopy, filename);
      strcat(safetycopy, "~");
    }
  }
  if(!useExFilename){
  if((sgf = fopen(FILENAME, "r")) != 0){
    fclose(sgf);
    strcpy(filename, FILENAME);
    if(forceOverwriteAndMakeSafetyCopy){
      if((sgf = fopen(FILENAME, "w"))==0){
	fprintf(stderr, "Bad file error!!\n");
	exit(1);
      }
    }
    else return false;
  }
  else if((sgf = fopen(FILENAME, "w"))==0){
   if((sgf = fopen(ALTERNATEFILE, "r")) != 0){
    fclose(sgf);
    strcpy(filename, ALTERNATEFILE);
    if(forceOverwriteAndMakeSafetyCopy){
      if((sgf = fopen(ALTERNATEFILE, "w"))==0){
	fprintf(stderr, "Bad file error!!\n");
	exit(1);
      }
    }
    else return false;
  }
   else if((sgf = fopen(ALTERNATEFILE, "w"))==0) {
     if((sgf = fopen(THIRDFILE, "r")) != 0){
       fclose(sgf);
       strcpy(filename, THIRDFILE);
       if(forceOverwriteAndMakeSafetyCopy){
	 if((sgf = fopen(THIRDFILE, "w"))==0){
	   fprintf(stderr, "Bad file error!!\n");
	   exit(1);
	 }
       }
       else return false;
     }
     else if((sgf=fopen(THIRDFILE,"w"))==0){
       fprintf(stderr, "Could not open file!\n");
       exit(1);
     }

   }
  }
  }
  fprintf(sgf, FILESTART);
  
  if(size > 9){
    stri[0] = size / 10 + '0';
    stri[1] = size %10 + '0';
    stri[2] = 0;
  }
  else{
    stri[0] = size + '0';
    stri[1] = 0;
  }
  
  fprintf(sgf, stri);
  fprintf(sgf, "]HA[");

  stri[0] = handi + '0';
  stri[1] = 0;

  fprintf(sgf, stri);
  fprintf(sgf, "]KM[");
  fprintf(sgf, "%c", (komi+'0'));
  //  fprintf(stderr, "%d\n", komi);
  if(plusFive) fprintf(sgf, ".5");
  fprintf(sgf,"]\nPW[");
  fprintf(sgf, plWhite);
  fprintf(sgf, "]\nPB[");
  fprintf(sgf, plBlack);
  fprintf(sgf, "]\nGN[");
  fprintf(sgf, gameName);
  fprintf(sgf, "]\nDT[");

  fprintf(sgf,"%d-%d-%d]\n", 1900+day->tm_year, 1+day->tm_mon,day->tm_mday);
  fprintf(sgf,FILEMED);

  if(handi > 0){
    fprintf(sgf, "AB");
    for(int i=0; i<handi; i++)
      fprintf(sgf, "[%c%c]", handix[i]+'a', handiy[i]+'a');
    fprintf(sgf, "\n");
  }
  
  
  writeFile(sgf, zero->getFirstNext());

  fprintf(sgf, FILEEND);
  fclose(sgf);

#if defined MAKESAFETYCOPY
  MakeSafetyCopy--;
  if(MakeSafetyCopy <= 0 &&
     forceOverwriteAndMakeSafetyCopy && safetycopy[0]!=0){
    MakeSafetyCopy = MAKESAFETYCOPY;
    system(safetycopy);
  }
#endif
  return true;
}

void GoFile::writeFile(FILE* sgf, Move* first){
  char* move;
  move = new char[10];
  int pl;
  while(first != 0){
    getMoveString(&move, first);
    if(first->nextL == 0){
      fprintf(sgf, move);
      first = first->getFirstNext();
    }
    else{
      while(first !=0 ){
	getMoveString(&move, first);
	fprintf(sgf, "(");
	fprintf(sgf, move);
	writeFile(sgf,first->getFirstNext());
	fprintf(sgf, ")\n");
	first = first->nextL;
      }
    }
  }
}

void GoFile::getMoveString(char ** move, Move* first){
  if(first->comment!=0){
    delete (*move);
    (*move) = new char[20+strlen(first->comment)];
  }
  (*move)[0]=';';
  (*move)[1] = (first->White) ? 'W' : 'B';
  (*move)[2] = '[';
  if((first->what &PASS) != PASS){
    (*move)[3]=first->x+'a';
    (*move)[4]=first->y+'a';
    (*move)[5]=0;
  }
  else (*move)[3] = 0;
  strcat(*move, "]");
  if(first->comment!=0){
    char CH[2];
    CH[1]=0;
    strcat(*move, "C[\n");
    for (int i=0; i<strlen(first->comment); i++){
      CH[0] = (first->comment)[i];
      if(CH[0] == '\\' || CH[0] == ']')
	strcat(*move, "\\");
      strcat(*move, CH);
    }
    strcat(*move, "\n]");
  }
  else strcat(*move, "\n");
  
}



Move* GoFile::parseFile(int& boardsz, int& handicap, int& komi, 
			bool& plusfive, int* handix, int*handiy){
  FILE* sgf;
  Move *zero;
  char ch;
  size = boardsz;
  if(filename[0] == 0){
    fprintf(stderr, "Bad filename error!\n");
    exit(1);
  }
  if ((sgf=fopen(filename, "r"))==0){
    return 0;
  }

  gotoNext(sgf, '(');
  gotoNext(sgf, ';');

  bool preWord = false;

  zero = new Move(0,0,true, OK);
  //  fprintf(stderr,"\nrecursive: size %d\n", size);

  Move* current = zero;

  // init -- find size and handicap
  while((ch = fgetchar(sgf))!= ';' && ch != '(' && ch != ')' && ch != EOF){
    switch(ch){
    case '[':
      gotoNext(sgf, ']');
      preWord = false;
      break;
    case 'P':
    case 'p':
      if(preWord) break;
      ch=fgetchar(sgf);
      if(ch == 'W' || ch == 'w'){
	fgetchar(sgf);
	parseString(sgf, plWhite);
	preWord = false;
      }
      else if(ch == 'B' || ch == 'b'){
	fgetchar(sgf);
	parseString(sgf, plBlack);
	preWord = false;
      }
      else if(ch == 'L' || ch == 'l'){
	if(fgetchar(sgf) != '['){
	  fprintf(stderr, "PL makes no sens in input!\n");
	  exit(1);
	  }
	ch = fgetchar(sgf);
	// adjust color of last move to the wish who may move next.
	if( ((ch == 'W' || ch == 'w') && current->White == WHITE) ||
	    ( ch != 'W' && ch != 'w' && current->White == BLACK))
	  {
	    Move* inb = current->prev;
	    
	    while(inb->White == current->White && inb->prev != 0)
	      inb = inb->prev;
	    
	    if(inb->prev != 0 && inb->White != current->White){
	      (inb->prev)->setNext(inb->getNext());

	      (inb->getNext())->prev = inb->prev;
	      inb->prev = current;

	      inb->setNext( current->getNext());
	      current->setNext(inb);
	      current = inb;
	    }
	    else{
	      current->nextAppend(-1,-1,!current->White, PASS);
	      current = current->getNext();
	    }
	    if(current->prev->comment !=0) {
	      current->comment = (current->prev)->comment; 
	      (current->prev)->comment = 0;
	    }
	  }
	fgetchar(sgf);
	preWord = false;
      }
      else preWord = true;
      break;
    case 'G':
    case 'g':
      if(preWord) break;
      ch=fgetchar(sgf);
      if(ch == 'N' || ch == 'n'){
	fgetchar(sgf);
	parseString(sgf, gameName);
	preWord = false;
      }
      else preWord = true;
      break;
    case 'H':
    case 'h':
      if(preWord) break;
      ch = fgetchar(sgf);
      if(ch == 'A' || ch == 'a'){
	ch = fgetchar(sgf);
	if(ch == '['){
	  handicap = fgetchar(sgf) - '0';
	  handi = handicap;
	  gotoNext(sgf, ']');
	  preWord = false;
	}
	else preWord = true;
      }
      break;
    case 'S':
    case 's':
      if(preWord) break;
      ch = fgetchar(sgf);
      if(ch == 'Z' || ch == 'z'){
	ch = fgetchar(sgf);
	if(ch == '['){
	  size = fgetchar(sgf) - '0';
	  ch = fgetchar(sgf);
	  if(ch != ']') {
	    size*=10;
	    size+= ch-'0';
	    gotoNext(sgf, ']');
	    preWord = false;
	  }
	  boardsz = size;
	}
	else preWord = true;
      }
      break;
    case 'A':
    case 'a':
      if(preWord) {
	break;
      }
      ch = fgetchar(sgf);
      if(ch == 'B' || ch == 'b'){
	for (int i=0; i<handicap; i++){
	  if(fgetchar(sgf) != '['){
	    fprintf(stderr, "HA <-> AB conflict in input file!\n");
	    exit(1);
	  }
	  handix[i] = fgetchar(sgf)-'a';
	  handiy[i] = fgetchar(sgf)-'a';    
	  gotoNext(sgf, ']');
	  preWord = false;
	}
 	if( fgetchar(sgf) == '[') {
 	  current=getInitialMoves(sgf, current, BLACK); 
 	  preWord = false;
	}
	fseek (sgf, -1, SEEK_CUR);
      }
      else if(ch == 'W' || ch == 'w'){
	if(fgetchar(sgf) != '['){
	  fprintf(stderr, "AW inconsistent in input file!\n");
	  exit(1);
	}
	current = getInitialMoves(sgf, current, WHITE);
	fseek (sgf, -1, SEEK_CUR);
	preWord = false;
      }
      else preWord = true;
      break;
    case 'K':
    case 'k':
      if(preWord) break;
      ch = fgetchar(sgf);
      if(ch == 'M' || ch == 'm'){
	if(fgetchar(sgf) != '['){
	  fprintf(stderr, "KM without value for komi in input file!\n");
	  exit(1);
	}
	komi = fgetchar(sgf)-'0';
	ch = fgetchar(sgf);
	if(ch == '.') {
	  if (fgetchar(sgf)== '0') plusfive = false;
	  else plusfive = true;
	}
	else plusfive = false;
	if(ch != ']') gotoNext(sgf, ']');
	preWord = false;
      }
      else preWord = true;
      break;

    case 'C':
    case 'c':
      char* comm;
      if((ch=fgetchar(sgf)) == '[') {
	long position = ftell(sgf);
	gotoNext(sgf,']');	
	int length = (int)(ftell(sgf) - position);
	comm = new char[length+1];
	fseek(sgf, position, SEEK_SET);
	bool lastBS = false;
	int i=0;
	ch=fgetchar(sgf);
	while((ch!=']' || lastBS) && ch!= EOF){
	  if(ch=='\\' && !lastBS) lastBS = true;
	  else {
	    lastBS=false;
	    comm[i] = ch; 
	    i++;
	  }
	  ch=fgetc(sgf);
	  while(ch == '\r') ch = fgetc(sgf);
	  /*	  if( !lastBS){
	    comm[i] = ch; 
	    i++;
	    }*/
	}
	comm[i-1]=0;
	if(comm[i-2]=='\n') comm[i-2]=0;
	current->setComm(comm);
      }
      break;

    default: preWord = true;
    }

  }

  if(ch == '(')                  // hack to enable variations before first move
    while(ch!= ')'){
      gotoNext(sgf, ';');
      ch = parseAppend(sgf, current);
      ch = gotoEither(sgf, '(',')');
    }

  else if(ch != ')'){
    parseAppend(sgf, current);
    zero->setLongestNext();
  }
  
  fclose(sgf);
  return zero;
}

Move* GoFile::getInitialMoves(FILE* sgf, Move* cur, bool white){
  char ch;
  char* comm=0;
  int x, y;

  //  gotoNext(sgf,'[');

  ch = fgetchar(sgf);
  if(ch!= ']'){
    x = ch-'a';
    y = fgetchar(sgf)-'a';
    if(y!=']' - 'a') ch = fgetchar(sgf);
  }
  else x = -1;

  if(ch != ']') {
    fprintf(stderr, "error with initially set stones!\n");
    exit(1);
  }


  if( x< 0 || x>= size || y<0 || y>=size){
    //    printf("fail %c\n",ch);
    cur->nextAppend(-1,-1,white, PASS,comm);
  }
  else{
    cur->nextAppend(x,y,white, OK, comm);
    //    fprintf(stderr, "%d, %d\n", x, y);
  }
  cur = cur->getNext();

  if(fgetchar(sgf) == '[')
    return getInitialMoves(sgf, cur, white);

  return cur;

}

char GoFile::parseAppend(FILE* sgf, Move* cur){
  char ch;
  char* comm=0;
  bool white = false;
  int x, y;
  if((ch = fgetchar(sgf))!= 'W' && ch != 'B' && ch != 'w' && ch!= 'b') {
    return ch;
  }
  if(ch == 'W' || ch == 'w') white = WHITE;
  else white = BLACK;
  gotoNext(sgf,'[');
  ch = fgetchar(sgf);
  if(ch!= ']'){
    x = ch-'a';
    if(ch!=']') {
      y = fgetchar(sgf)-'a';
      if(y!=']' - 'a') ch = fgetchar(sgf);
      //    printf("\nsucc %d, %d\n",x,y);
    }}
  else x = -1;
  bool preWord =false;
  while((ch = fgetchar(sgf))!= ';' && ch != ')' && ch !='(' && ch != EOF){
    switch(ch){
    case '[':
      gotoNext(sgf, ']');
      preWord = false;
      break;
    case 'C':
    case 'c':
      if((ch=fgetchar(sgf)) == '[') {
	long current = ftell(sgf);
	gotoNext(sgf,']');	
	int length = (int)(ftell(sgf) - current);
	if(comm!=0) delete comm;
	comm = new char[length+1];
	fseek(sgf, current, SEEK_SET);
	bool lastBS = false;
	int i=0;
	ch=fgetchar(sgf);
	while((ch!=']' || lastBS) && ch!= EOF){
	  if(ch=='\\' && !lastBS) lastBS = true;
	  else {
	    lastBS=false;
	    comm[i] = ch; 
	    i++;
	  }
	  ch=fgetc(sgf);
	  while(ch == '\r') ch = fgetc(sgf);
	  /*	  if( !lastBS){
	    comm[i] = ch; 
	    i++;
	    }*/
	}
	comm[i-1]=0;
	if(comm[i-2]=='\n') comm[i-2]=0;
      }
      break;
    default:
      preWord = true;
    }
  }
  if( x< 0 || x>= size || y<0 || y>=size){
    //    printf("fail %c\n",ch);
    cur->nextAppend(-1,-1,white, PASS,comm);
  }
  else{
    cur->nextAppend(x,y,white, OK, comm);
    //    fprintf(stderr, "%d, %d\n", x, y);
  }
  cur = cur->getNext();
  if(ch!=';' && ch != ')' && ch !='(') ch = gotoEither(sgf,';',')','(');
  if(ch == ')') return ch;
  if(ch == ';') return parseAppend(sgf, cur);
  if(ch == '(') {
    while(ch!= ')'){
      gotoNext(sgf, ';');
      ch = parseAppend(sgf, cur);
      ch = gotoEither(sgf, '(',')');
    }
    return ')';
  }
}


void GoFile::gotoNext(FILE* sgf, char ch){
  char c='?';
  bool lastBS = false;
  while((c!=ch || lastBS) && c!= EOF){
    if(c=='\\' && !lastBS) lastBS = true;
    else lastBS=false;
    c=fgetc(sgf);
  }

  if(c==EOF){
    fprintf(stderr, "End of File reached unexpectedly!\n");
    exit(1);
  }
}

char GoFile::gotoEither(FILE* sgf, char ch, char ch2){
  char c;
  while((c=fgetc(sgf))!= ch && c != ch2 && c!= EOF) 
    if(c=='[') gotoNext(sgf, ']');
  if(c==EOF){
    fprintf(stderr, "End of File reached unexpectedly!\n");
    exit(1);
  }
  return c;
}

char GoFile::gotoEither(FILE* sgf, char ch, char ch2, char ch3){
  char c;
  while((c=fgetc(sgf))!= ch && c != ch2 && c!= ch3 && c!= EOF)
    if(c=='[') gotoNext(sgf, ']');

  if(c==EOF){
    fprintf(stderr, "End of File reached unexpectedly!\n");
    exit(1);
  }
  return c;
}

void GoFile::parseString(FILE* sgf, char* string){
  char ch = ' ';
  int i;
  for(i=0; i<MAXSTRING && ch != ']' && ch != EOF; i++)
    string[i] = (ch = fgetchar(sgf));
  string[i-1] = 0;
  if(ch == EOF){
    fprintf(stderr, "Bad input file!\n");
    exit(1);
  }
}
char GoFile::fgetchar(FILE* file){
  char ch;
  while(((ch=fgetc(file)) == ' ' || ch == '\n' || ch == '\r') && ch != EOF);
  //  fprintf(stderr, "%c",ch);
  return ch;
}
