/*
// the defines are in GoDefines.cc
#include "GoDefines.cc"

// now the program starts
#include <stdio.h>
#include <stdlib.h>
//#include <iostream.h>
#include <string.h>
#include <new.h>
#include <unistd.h>
#include <time.h>
*/
// you need to port all the public functions of GoGraphic in order to port 
// goVRec to another system (but that's it...)

#ifdef USE_QT
  #undef QTFONTTEST
  #include <qapplication.h>
  #include <qwidget.h>
  #include <qevent.h>
  #include <qpainter.h>
  #include <qcolor.h>
  #include <qpen.h>
  #include <qfontmetrics.h>
  #ifdef QTFONTTEST
    #include <qfontdialog.h>
    #include <qfontdatabase.h>
    #include <qstringlist.h>
  #endif

#else
  #include <X11/Xlib.h>
  #include <X11/keysym.h>
  #include <X11/Xutil.h>
//#include <X11/Xatom.h>
#endif


#define EXPO 1
#define BUTTON 2
#define KEY 3
#define RESIZE 4
#define MID_BUTTON 5
#define RIGHT_BUTTON 6

#define PgUP 1
#define PgDOWN 2
#define LEFT 3
#define RIGHT 4
#define QUITKEY 5
#define YESKEY 6
#define NOKEY 7

class GoEvent {
public:
  int type;
  int xtype;
  int y;
  bool shift;
#ifdef USE_QT
  GoEvent::GoEvent() { nextEvent=NULL; }
private:
  GoEvent* nextEvent;
friend class GoWidget;
friend class GoGraphic;
#endif
};

#ifdef USE_QT
class GoWidget : public QWidget
{
public:
  GoWidget(QWidget* parent=0, const char* name=0, int f=0);
  void paintEvent(QPaintEvent* e);
  void mousePressEvent(QMouseEvent* e);
  void keyPressEvent(QKeyEvent* e);
  void resizeEvent(QResizeEvent *e);
  void closeEvent(QCloseEvent *e);
private:
  GoEvent* storedEvent;
  void storeEvent(GoEvent* ev);
friend class GoGraphic;
};

GoWidget::GoWidget(QWidget* parent, const char* name, int f) :
  QWidget(parent, name, f)
{
  storedEvent = NULL;
}
#endif

class GoGraphic {
private:
#ifdef USE_QT
  GoWidget* qtWidget;
  QApplication* qtApp;
  QPainter* p;
  static QPen blackpen, whitepen, backpen;
  static QBrush blackbrush, whitebrush, backbrush;
  static QFont nQFont, sQFont;
#else
  Display *display;
  Window window;
  GC blackgc, whitegc, backgc, smallgc;
#endif
  int s_height, s_width, s_ascent, s_descent;
#ifdef USE_QT
#else
  XKeyboardState kbstate;
  XEvent event;
  KeySym key;
  void resetKeyboard();
  GC CreateGC(Display *display, Drawable drawable, 
	      unsigned long forecolor, unsigned long backcolor);
#endif


public:
  GoGraphic();
  void setupWindow(char*, int, char**);
  void setSizes(int*, int*);
  //  void doDraw();
  void point(int x, int y, bool black = BLACK);
  void line(int,int,int,int, bool black = BLACK);
  void rectangle(int,int,int,int, bool black = BLACK);
  void fRectangle(int,int,int,int, bool);
  void emptyRectangle(int,int,int,int);
  void backRectangle(int,int,int,int);
  void circle (int x_tl, int y_tl, int rad, bool black, bool back = false);
  void fCircle (int x_tl, int y_tl, int rad, bool black, bool back = false);
  void playsound(int);
  void dString(int,int,const char*);
  void dSString(int,int,const char*);
  void dOneLine(int,int);
  void getEvent(GoEvent*);
};

void GoGraphic::getEvent(GoEvent* ev){
#ifdef USE_QT
  while(qtWidget->storedEvent==NULL){
    usleep(20000);	// maybe not very portable
	 qtApp->processEvents();
  }
  GoEvent* oldest = qtWidget->storedEvent;
  ev->type  = oldest->type;
  ev->xtype = oldest->xtype;
  ev->y     = oldest->y;
  ev->shift = oldest->shift;
  qtWidget->storedEvent = oldest->nextEvent;
  delete oldest;
#ifdef QTFONTTEST
  if((ev->type == BUTTON) && (ev->y > 200)){
    bool qfok;
    QFont qf = QFontDialog::getFont(&qfok,qtWidget,"Schriftartwahl");
	 if(qfok) {nQFont=qf; sQFont=qf;}
	 else exit(2);
/*
    QFont::initialize();
	 QStringList qsl = (new QFontDatabase())->families(FALSE);
	 QStringList::Iterator it;
	 FILE* dat = fopen("qtfonts.txt","w");
	 for( it = qsl.begin(); it != qsl.end(); ++it ){
      fprintf(dat,"%s\n",(*it).latin1());
    }
	 fclose(dat);
*/
  }
#endif
#else
  XNextEvent(display,&event);
  switch(event.type){
  case ButtonPress:
    ev->type = BUTTON;
    break;
  case KeyPress:
    ev->type = KEY;
    break;
  case Expose:
    ev->type = EXPO;
    return;
  case ConfigureNotify:
    ev->type = RESIZE;
    ev->xtype= ((XConfigureEvent*)&event)->width;
    ev->y    = ((XConfigureEvent*)&event)->height;
    //    XResizeWindow(display, window, ev->xtype, ev->y);
    return;
  //    printf("%d,%d\n", ((XResizeRequestEvent*)&event)->width, 
  //	   ((XResizeRequestEvent*)&event)->height);

  default:
    ev->type = NOTHING;
    return;
  }
  if(ev->type == BUTTON){
    ev->xtype = event.xbutton.x;
    ev->y = event.xbutton.y;
    if(event.xbutton.button == Button2) ev->type = MID_BUTTON;
    else if(event.xbutton.button == Button3) ev->type=RIGHT_BUTTON;
    return;
  }
  ev->shift = ((XKeyEvent*)&event)->state&ShiftMask;
  key = XLookupKeysym((XKeyEvent*)&event, 0);
  switch(key){
  case XK_Page_Up:
    ev->xtype = PgUP;
    break;
  case XK_Page_Down:
    ev->xtype = PgDOWN;
    break;
  case XK_Right:
    ev->xtype = RIGHT;
    break;
  case XK_Left:
    ev->xtype = LEFT;
    break;
  case XK_Q:
  case XK_q:
    ev->xtype = QUITKEY;
    break;
  case XK_y:
  case XK_Y:
    ev->xtype = YESKEY;
    break;
  case XK_n:
  case XK_N:
    ev->xtype = NOKEY;
    break;
  default:
    ev->xtype = NOTHING;
  }
#endif
}

#ifdef USE_QT
void GoWidget::keyPressEvent ( QKeyEvent * e )
{
  GoEvent* ev = new GoEvent;
  switch((Key)(e->key())){
  case Key_PageUp:
    ev->xtype = PgUP;
    break;
  case Key_PageDown:
    ev->xtype = PgDOWN;
    break;
  case Key_Right:
    ev->xtype = RIGHT;
    break;
  case Key_Left:
    ev->xtype = LEFT;
    break;
  case Key_Q:
    ev->xtype = QUITKEY;
    break;
  case Key_Y:
    ev->xtype = YESKEY;
    break;
  case Key_N:
    ev->xtype = NOKEY;
    break;
  default:
    ev->xtype = NOTHING;
    e->ignore();
  }
  ev->type = KEY;
  ev->shift = e->state()&ShiftButton;
  storeEvent(ev);
}

void GoWidget::mousePressEvent(QMouseEvent* e)
{
  GoEvent* ev = new GoEvent;
  ev->xtype = e->x();
  ev->y = e->y();
  switch(e->button()){
  case MidButton:
    ev->type = MID_BUTTON;
    break;
  case RightButton:
    ev->type = RIGHT_BUTTON;
    break;
  default:
    ev->type = BUTTON;
  }
  storeEvent(ev);
}

void GoWidget::resizeEvent(QResizeEvent *e)
{
  GoEvent* ev = new GoEvent;
  ev->type = RESIZE;
  ev->xtype= e->size().width();
  ev->y    = e->size().height();
  /* XXX http://doc.trolltech.com/2.3/qwidget.html#7d375f Achtung Redundanzgefahr: Ein neu_Zeichnen-Ereignis wird automatisch extra geliefert */
  storeEvent(ev);
}

void GoWidget::closeEvent(QCloseEvent *e)
{
  GoEvent* ev = new GoEvent;
  ev->type = KEY;
  ev->xtype = QUITKEY;
  e->ignore();
  storeEvent(ev);
}

void GoWidget::paintEvent(QPaintEvent* e)
{
  GoEvent* ev = new GoEvent;
  ev->type = EXPO;
  storeEvent(ev);
}

void GoWidget::storeEvent(GoEvent* ev)
{
  if(storedEvent==NULL) storedEvent = ev;
  else{
    GoEvent* last = storedEvent;
	 while(last->nextEvent != NULL)
	   last = last->nextEvent;
	 last->nextEvent = ev;
  }
}
#endif

void GoGraphic::setSizes(int* w, int*h){
  *w=s_width;
  *h=s_height;
}

void GoGraphic::dString (int x, int y, const char* str){
#ifdef USE_QT
  p->setFont(nQFont);
  p->setPen(blackpen);
  p->drawText( x, y, str, strlen(str) );
#else
  XDrawString(display, window, blackgc, 
	      x,y, str, strlen(str));
#endif
}

void GoGraphic::dSString (int x, int y, const char* str){
#ifdef USE_QT
  p->setFont(sQFont);
  p->setPen(blackpen);
  p->drawText( x, y, str, strlen(str) );
#else
  XDrawString(display, window, smallgc, 
	      x,y, str, strlen(str));
#endif
}

void GoGraphic::dOneLine (int x, int y){
  line(x-2, y-s_ascent, x-2, y-s_descent);
}

void GoGraphic::point(int x, int y, bool black){
#ifdef USE_QT
  p->setPen((black==BLACK)?blackpen : whitepen);
  p->drawPoint(x, y);
#else
  XDrawPoint(display,window,(black==BLACK)?blackgc : whitegc, x, y);
#endif
}

void GoGraphic::line (int x1, int y1, int x2, int y2, bool black){
#ifdef USE_QT
  p->setPen((black==BLACK)?blackpen : whitepen);
  p->drawLine(x1, y1, x2, y2);
#else
  XDrawLine(display, window, (black==BLACK) ? blackgc : whitegc, x1, y1, x2, y2);
#endif
}
void GoGraphic::rectangle(int x, int y, int width, int height, 
			  bool black){
#ifdef USE_QT
  p->setPen((black==BLACK)?blackpen : whitepen);
  p->setBrush(Qt::NoBrush);
  p->drawRect(x,y,width,height);
#else
  XDrawRectangle(display,window, (black==BLACK)?blackgc:whitegc, 
		 x,y,width,height);
#endif
}

void GoGraphic::fRectangle (int x, int y, int width, int height, bool black){
#ifdef USE_QT
  p->setPen(Qt::NoPen);
  p->setBrush((black==BLACK)?blackbrush: whitebrush);
  p->drawRect(x,y,width,height);
#else
  XFillRectangle(display,window, (black == BLACK)? blackgc : whitegc, 
		 x,y,width,height);
#endif
}

void GoGraphic::emptyRectangle (int x, int y, int width, int height){
#ifdef USE_QT
  p->setPen(Qt::NoPen);
  p->setBrush(whitebrush);
  p->drawRect(x,y,width,height);
#else
  XFillRectangle(display,window, whitegc, x,y,width,height);
#endif
}

void GoGraphic::backRectangle (int x, int y, int width, int height){
#ifdef USE_QT
  p->setPen(Qt::NoPen);
  p->setBrush(backbrush);
  p->drawRect(x,y,width,height);
#else
  XFillRectangle(display,window, backgc, x,y,width,height);
#endif
}

void GoGraphic::circle (int x_tl, int y_tl, int rad, bool black = BLACK, 
			bool back){
#ifdef USE_QT
  p->setPen((back)?backpen : ((black==BLACK)?blackpen : whitepen));
  p->setBrush(Qt::NoBrush);
  p->drawEllipse( x_tl, y_tl, rad, rad );
#else
  XDrawArc(display, window, 
	   (back)?backgc : ((black==BLACK)?blackgc : whitegc), 
	   x_tl, y_tl, rad, rad, 0, 360*64);
#endif
}

void GoGraphic::fCircle (int x_tl, int y_tl, int rad, bool black,
			 bool back){
#ifdef USE_QT
  p->setPen(Qt::NoPen);
  p->setBrush((back)?backbrush : ((black==BLACK)?blackbrush : whitebrush));
  p->drawEllipse( x_tl, y_tl, rad, rad );
#else
  XFillArc(display, window, 
	   (back)?backgc : ((black==BLACK)?blackgc : whitegc), 
	   x_tl, y_tl, rad, rad, 0, 360*64);
#endif
}

void GoGraphic::playsound(int wsound){
#ifdef USE_QT
  #warning Sound implementation is still primitive
  qtApp->beep();
#else
  XKeyboardControl vals;
  
  if(wsound == STONETONE) vals.bell_pitch = STONEPITCH;
  else if (wsound == SYSTEMTONE) vals.bell_pitch = SYSTEMPITCH;
  else vals.bell_pitch = ALARMPITCH;
  vals.bell_duration = SOUNDDURATION;
  vals.bell_percent = SOUNDVOLUME;
  
  XChangeKeyboardControl(display,
                         (KBBellPitch | KBBellDuration | KBBellPercent),
                         &vals);
  XBell(display, 100);
  XFlush(display);
  resetKeyboard();
#endif
}

#ifdef USE_QT

  /* XXX GGF. SPÄTER IMPLEMENTIEREN, WENN SOUND IMPLEMENTIERT */

#else
void GoGraphic::resetKeyboard(){
  XKeyboardControl vals;
  
  
  /* Return display's bell pitch/duration to normal! */
  
  vals.bell_pitch = kbstate.bell_pitch;
  vals.bell_duration = kbstate.bell_duration;
  vals.bell_percent = kbstate.bell_percent;

  XChangeKeyboardControl(display, 
			 KBBellPitch | KBBellDuration|KBBellPercent, &vals); 
  XFlush(display);
}
#endif

//  void GoGraphic::doDraw(){
//    XFlush(display);
//  }

GoGraphic::GoGraphic(){
#ifdef USE_QT
#else
 /* Connect to X server: */
  
  display = XOpenDisplay(NULL);
  if (display == NULL)
    {
      fprintf(stderr, "Can't connect to display!\n");
      exit(1);
    }
  
  XGetKeyboardControl(display, &kbstate); 
#endif
}

#ifdef USE_QT
QFont GoGraphic::nQFont = QFont();
#ifdef SMALLFONT
QFont GoGraphic::sQFont = QFont(SMALLFONT);
#else
QFont GoGraphic::sQFont = nQFont;
#endif
QPen GoGraphic::blackpen = QPen(QColor(0,0,0));
QPen GoGraphic::whitepen = QPen(QColor(0xFF,0xFF,0xFF));
QPen GoGraphic::backpen  = QPen(QColor(REDCOLORVALUE,GREENCOLORVALUE,BLUECOLORVALUE));
QBrush GoGraphic::blackbrush = QBrush(QColor(0,0,0));
QBrush GoGraphic::whitebrush = QBrush(QColor(0xFF,0xFF,0xFF));
QBrush GoGraphic::backbrush  = QBrush(QColor(REDCOLORVALUE,GREENCOLORVALUE,BLUECOLORVALUE));
#endif

void GoGraphic::setupWindow(char* filename, int argc, char** argv) {
#ifdef USE_QT
  qtApp = new QApplication(argc, argv);
  qtWidget = new GoWidget(0);
  qtWidget->resize(WINDOWWIDTH, WINDOWHEIGHT);	// XXX redundante Ereignisse? http://doc.trolltech.com/2.3/qwidget.html#8fcbbe
//qtApp->setMainWidget(&qtWidget);
  qtWidget->show();

  p = new QPainter(qtWidget);
  p->setRasterOp(Qt::CopyROP);	/* ist wohl nicht nötig */
/* Qt siehe http://doc.trolltech.com/2.3/qfont.html, http://doc.trolltech.com/2.3/qfontmetrics.html#200b74; X siehe http://tronche.com/gui/x/xlib/graphics/font-metrics/#XFontStruct */
  s_width= qtWidget->fontMetrics().maxWidth();	/* XXX ist das richtig? Bild bei http://doc.trolltech.com/2.3/qfontmetrics.html#a6372a ist bedenklich */
  s_height=qtWidget->fontMetrics().height();
  s_ascent=qtWidget->fontMetrics().ascent();
  s_descent=qtWidget->fontMetrics().descent()+1;
#else
  Window rootwindow;
  int screen, black, white;
  XSetWindowAttributes attr;
  unsigned long attr_mask;

  screen = DefaultScreen(display);
  rootwindow = RootWindow(display, screen);
  
  /* Open window: */
  
  attr.event_mask = (StructureNotifyMask|KeyPressMask | KeyReleaseMask 
		     | ButtonPressMask |StructureNotifyMask|
                     ExposureMask | VisibilityChangeMask);
  attr.colormap = DefaultColormap(display, screen);
  attr.border_pixel = black;
  attr.background_pixel = white;
  attr.backing_store=Always;
  attr.save_under=true;
  attr_mask = ExposureMask|StructureNotifyMask|CWEventMask | CWBackPixel | CWBorderPixel|CWColormap;
  
  window = XCreateWindow(display, rootwindow, 0, 0, 
			 windowwidth, windowheight, 0,
                         DefaultDepthOfScreen(DefaultScreenOfDisplay(display)),
                        InputOutput, DefaultVisual(display, screen),
                         attr_mask, &attr);
  
  char windowname[MAXSTRING+1];
  
  strcpy(windowname, "goVRec");
  
  if(strlen(filename)>0){
    strcat(windowname, " ");
    strcat(windowname, filename);
    if(strcmp(".sgf",&windowname[strlen(windowname) -4])==0) 
      windowname[strlen(windowname)-4] = 0;
  }
  XStoreName(display, window, windowname);

  /* Get primitive colors: */
  
  black = BlackPixel(display, screen);
  white = WhitePixel(display, screen);
  
  /* Create primitive graphics contexts: */

  XColor real;
  real.pixel=(REDCOLORVALUE<<16)+(GREENCOLORVALUE<<8) + BLUECOLORVALUE;

  blackgc = CreateGC(display, window, black, black);
  smallgc = CreateGC(display, window, black, black);
  whitegc = CreateGC(display, window, white, black); 
  backgc = CreateGC(display,window, 
		    (REDCOLORVALUE<<16)+(GREENCOLORVALUE<<8) + BLUECOLORVALUE,
		    black); 

#if defined SMALLFONT
  XGCValues xgcvalues;

  //  XGetGCValues(display, blackgc, GCFont, &xgcvalues);
  Font small = XLoadFont(display, SMALLFONT);
  xgcvalues.font=small;
  XChangeGC(display, smallgc, GCFont, &xgcvalues);
#endif
  XFontStruct* fontStr;
  fontStr=XQueryFont(display, XGContextFromGC(smallgc));
  //  char nam[80];
  //    XGetFontProperty(fontStr, XA_FONT_NAME, (unsigned long*) nam);
  s_width= (fontStr->max_bounds).width;
  s_height=(fontStr->max_bounds).ascent+(fontStr->max_bounds).descent;
  s_ascent=(fontStr->max_bounds).ascent;
  s_descent=(fontStr->max_bounds).descent;
  //  printf("%d,%d\n", s_width, s_height);

  

  /* Bring window up! */

  
  
  XMapWindow(display, window);
  XMapRaised(display, window);

  // Wait for the MapNotify event & Expose event

  XEvent e;
  XNextEvent(display, &e);
  while(e.type != MapNotify) XNextEvent(display, &e);
  while(e.type != Expose) XNextEvent(display, &e);
  // apparently, we get a configureNotify in the beginning
  //  while(e.type != ConfigureNotify) XNextEvent(display, &e);

  // now set the input hint to get listed in the window list
  // which does not work at the moment.

  XWMHints wmhints;
  wmhints.input = true;
  wmhints.flags |= InputHint;
  XSetWMHints (display, window, &wmhints);

  XFlush(display);

  /*
  XGCValues xgcvalues;

  //  XGetGCValues(display, blackgc, GCFont, &xgcvalues);
  xgcvalues.font=small;
  XChangeGC(display,blackgc, GCFont, &xgcvalues);

  XFontStruct* fontStr;
  fontStr=XQueryFont(display, XGContextFromGC(blackgc));
  //  char nam[80];
  //    XGetFontProperty(fontStr, XA_FONT_NAME, (unsigned long*) nam);
  if(fontStr!=0) printf("%d, %d, %d \n", (fontStr->max_bounds).width,
			(fontStr->max_bounds).ascent,
			(fontStr->max_bounds).descent);
  else printf("Font didn't exist\n");

  
  char ** names;
  int howmany;
  names= XListFonts(display, "*", 100,&howmany);
  printf("%d\n",howmany);
  for(int i=0; i<howmany; i++)
    printf("%s\n",names[i]);
  */
#endif
} 

#ifdef USE_QT
#else
/* Create a graphics context: */

GC GoGraphic::CreateGC(Display *display, Drawable drawable, 
		       unsigned long forecolor, unsigned long backcolor)
{
  XGCValues xgcvalues;
  GC gc;
  
  xgcvalues.foreground = forecolor;
  xgcvalues.background = backcolor;
  gc = XCreateGC(display,drawable,(GCForeground | GCBackground),
                 &xgcvalues);


  return(gc);
}
#endif

