/*
govrec.cc
A go recording program for the agenda VR3 (and possibly other Linux PDAs)
Andreas Balser
 */

// the defines are in GoDefines.cc
#include "GGoDefines.cc"

// now the program starts
#include <stdio.h>
#include <stdlib.h>
//#include <iostream.h>
#include <string.h>
#include <new>
#include <unistd.h>
#include <time.h>
int windowheight, windowwidth;
#include "GoGraphic.cc"
#include "MoveL.cc"
#include "GoBoxes.cc"
#include "GoBoard.cc"
//#include "MoveSelectButtons.cc"
#include "GoFile.cc"
#include "GoControl.cc"
void speicherMangel();


int main(int nargs, char** args){
  // just to pass the first event from init to gameplay
  bool checkEvent = true; 
  GoEvent event;
  GoGraphic* graph;
  GoControl* control;

  std::set_new_handler(speicherMangel);

  windowwidth = WINDOWWIDTH;
  windowheight = WINDOWHEIGHT;
  
  graph = new GoGraphic();
  control = new GoControl(nargs, args, graph);

  if(!control->gameStarted()){
    control->initialiseNewGame();
    //    graph->doDraw();
    while(checkEvent){
      graph->getEvent(&event);
      switch(event.type){
      case BUTTON:
	if(event.xtype >=windowwidth-BUTTONSIZE-1
	   || event.y >= windowheight-windowwidth){
	  checkEvent = false;
	  break;
	}
	control->handleInitClick(event.xtype, event.y);
	break;
      case RESIZE:
	windowwidth = event.xtype;
	windowheight = event.y;
	if (windowwidth > windowheight - 3*BUTTONSIZE - 2*FONTHEIGHT){
	  windowwidth = windowheight - 3*BUTTONSIZE - 2*FONTHEIGHT;
	graph->fRectangle (windowwidth,0,event.xtype - windowwidth, 
			       windowheight, BLACK);
	}
	control->resize();
	break;
      case EXPO:
	control->exposed();
	break;
      }
    }
    control->gameStart();
  }

  while(true){                                   // the actual game-play
    control->autoSave();
    if(checkEvent) graph->getEvent(&event);
    checkEvent = true;
    switch(event.type){
    case BUTTON:                             //MouseButton
      control->handleGameClick(event.xtype, event.y);
      
      break;

    case MID_BUTTON:
    case RIGHT_BUTTON:
      control->handleOtherClick(event.type);
      break;

    case KEY:                                       //KeyPress
      control->handleKeyPress(event);

      break;

    case RESIZE:
      windowwidth = event.xtype;
      windowheight = event.y;
      if (windowwidth > windowheight - 3*BUTTONSIZE - 2*FONTHEIGHT){
	windowwidth = windowheight - 3*BUTTONSIZE - 2*FONTHEIGHT;
	graph->fRectangle (windowwidth,0,event.xtype - windowwidth, 
			       windowheight, BLACK);
      }
      control->resize();
      break;
    
    case EXPO:
      control->exposed();
      break;
    }
  }
  exit(0);

}

void speicherMangel(){
  fprintf(stderr,"Memory problem!\n");
  exit(1);
}

